/*
 * Copyright 2011 Raster Software Vigo
 *
 * This file is part of ttos
 *
 * ttos is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * ttos is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

/*
 * clients.c
 *
 *  Created on: 18/09/2010
 *      Author: raster
 */

#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "widgets.h"
#include "clients.h"
#include "minibus/minibus.h"
#include "other.h"
#include "tabs.h"
#include "timer.h"
#include "sharemem.h"
#include "ipc.h"

/***
 * Creates a new _client_data object, and initializates it.
 *
 * @param connection The new connection stablished
 * @param data Private data to store in the client object
 *
 * @return A new _client_data object if success, or NULL on failure
 */

struct _client_data *clients_new(struct minibus_connection *connection,void *data) {

	struct _client_data *object;

	if (connection==NULL) {
		return (NULL);
	}

	object=(struct _client_data *)malloc(sizeof(struct _client_data));

	if (object==NULL) {
		return (NULL);
	}

	object->conn=connection;
	object->app_name=NULL;
	object->uid=minibus_get_uid(connection);
	object->data=data;
	object->window_manager=false;
	object->window=0; // by default, all pages go to window 0
	object->edit_tab=NULL;
	object->show_tab=NULL;
	minibus_list_init(&object->tabs,tabs_free_tab,NULL);
	minibus_list_init(&object->sharemem,sharemem_free,NULL);

	/*
	 *  The connection private data will be the client object.
	 * This allows to access this object during callbacks
	 */

	minibus_set_connection_data(connection,object);

	return (object);
}

/**
 * Destroys a _client_data object
 *
 * @param object The client_data object to be destroyed
 *
 */

void clients_destroy(void *data) {

	struct _client_data *object=(struct _client_data *)data;
	struct _timer *timer;
	struct minibus_list_element *element,*next,*element2,*next2;
	struct ipc_bus_element *ipc_element;
	struct ipc_subscriber *subscriber;

	if (!object)
		return;

	for(element=timers.list;element!=NULL;element=next) {
		next=element->next;
		timer=element->data;
		if (timer->conn==object->conn) {
			minibus_list_remove(element,true);
		}
	}

	for(element=ipc.list;element;element=next) {
		next=element->next;
		ipc_element=element->data;
		if (ipc_element->owner==object->ipc_uid) {
			minibus_list_remove(element,true);
			continue;
		}
		for(element2=ipc_element->subscribers.list;element2;element2=next2) {
			next2=element2->next;
			subscriber=element2->data;
			if (subscriber->ipc==object->ipc_uid) {
				minibus_list_remove(element2,true);
				break;
			}
		}
	}

	uid_release_uid(ttos_ipc_uid,object->ipc_uid);

	if (object->app_name) {
		free(object->app_name);
	}
	minibus_list_free(&object->tabs);
	minibus_list_free(&object->sharemem);

	free(object);
}

/**
 * Manages the message for initialization of a client. It allows a client to become window manager, and to know its UID.
 * It receives a long with the desired flags, and a string with the application name.
 * It sends a first message to the client which contains the approved flags, and the UID assigned to it. A second message
 * contains the current resolution and info about the pixel format
 *
 * @param port The port in which the message was received
 * @param answer_port The port to which send the errors
 * @param connection The client's connection
 * @param private_data Private data
 *
 * @return 0 on success; -1 on error
 */

uint32_t clients_init(uint32_t port, uint32_t answer_port, struct minibus_connection *connection, void *private_data) {

	uint32_t flags;
	int32_t error;
	Cstring app_name;
	struct minibus_data_element *data;
	struct _client_data *client;
	struct minibus_connection *wm_conn;

	data=minibus_read_params(connection,"lc",&error);

	if(!data) {
		switch(error) {
		case -1:
			other_send_error(connection,answer_port,TTOS_ERROR_NO_MEMORY);
			return -1;
		break;
		case -2:
			other_send_error(connection,answer_port,TTOS_ERROR_CANT_READ_DATA);
			return -1;
		break;
		case -3:
			other_send_error(connection,answer_port,TTOS_ERROR_INCORRECT_FORMAT);
			return -1;
		break;
		}
	}

	flags=data[0].value32;
	app_name=data[1].string;

	client=minibus_get_connection_data(connection);
	if (client->app_name) {
		free(client->app_name);
	}

	if (app_name[0]) {
		client->app_name=strdup(app_name);
	} else {
		client->app_name=strdup("");
	}
	free(data);

	/*
	 * Send a "new client" message to the window manager
	 */

	if (window_manager) {
		wm_conn=window_manager->conn;
		minibus_start_message(write_conn,TTOS_SYSTEM_PORT_2,TTOS_NO_ANSWER);
		minibus_write_long(write_conn,TTOS_SR_NEW_CLIENT);
		minibus_write_long(write_conn,client->uid);
		minibus_write_string(write_conn,client->app_name);
		minibus_server_send(write_conn,wm_conn,false);
	}

	if (flags&TTOS_INIT_ASK_WM) {
		if (!window_manager) {
			window_manager=minibus_get_connection_data(connection);
			window_manager->window_manager=true;
		} else {
			flags&=~TTOS_INIT_ASK_WM;
		}
	}

	client->ipc_uid=uid_get_uid(ttos_ipc_uid,TTOS_IPC_UID_MASK);

	// Return the new flags and the screen resolution

	minibus_start_message(write_conn,TTOS_SYSTEM_PORT_1,TTOS_NO_ANSWER);
	minibus_write_long(write_conn,TTOS_SR_INIT);
	// return the accepted flags
	minibus_write_long(write_conn,flags);
	// the assigned UID
	minibus_write_long(write_conn,minibus_get_uid(connection));
	minibus_write_long(write_conn,client->ipc_uid);
	// the pixel format
	minibus_write_long(write_conn,TTOS_PIXEL_FORMAT_TOMTOM_RGB565);
	minibus_write_byte(write_conn,2);      // bytes per pixel
	minibus_write_byte(write_conn,0);      // right shift for R
	minibus_write_byte(write_conn,8);      // left  shift for R
	minibus_write_long(write_conn,0xF800); //   bit mask  for R
	minibus_write_byte(write_conn,0);      // right shift for G
	minibus_write_byte(write_conn,3);      // left  shift for G
	minibus_write_long(write_conn,0x07E0); //   bit mask  for G
	minibus_write_byte(write_conn,3);      // right shift for B
	minibus_write_byte(write_conn,0);      // left  shift for B
	minibus_write_long(write_conn,0x001F); //   bit mask  for B
	minibus_write_byte(write_conn,0);      // right shift for A
	minibus_write_byte(write_conn,0);      // left  shift for A
	minibus_write_long(write_conn,0x0000); //   bit mask  for A; being 0 means it uses separated alpha
	minibus_server_send(write_conn,connection,false);
	clients_send_resolution(connection);
	other_send_error(connection,answer_port,TTOS_ERROR_NO_ERROR);
	return 0;
}


/**
 * Sends the current window resolution to a client
 *
 * @param connection The client's connection
 */

void clients_send_resolution(struct minibus_connection *conn) {

	struct _client_data *object=minibus_get_connection_data(conn);

	minibus_start_message(write_conn,TTOS_SYSTEM_PORT_1,TTOS_NO_ANSWER);
	minibus_write_long(write_conn,TTOS_SR_NEW_RESOLUTION);
	minibus_write_short(write_conn,windows[object->window].ox);
	minibus_write_short(write_conn,windows[object->window].oy);
	minibus_write_short(write_conn,windows[object->window].w);
	minibus_write_short(write_conn,windows[object->window].h);
	minibus_server_send(write_conn,conn,false);
}


/**
 * Defines a set of windows on screen, and sends, if required, a NEW_RESOLUTION message to each client. If a client had
 * assigned a window which dissapears in the new list, it will be relocated in the window 0
 *
 * The list of windows must be in the message. The first element, a LONG integer, must be the number of windows to be
 * defined. Following it will come as many tuples of four SHORT integers (X and Y coordinates, and Width and Height, for
 * each window) and one byte (0-> window is not visible, 1-> window is visible).
 *
 * If there's an error, it will send an error message; if not, it will send an OK message
 *
 * @param port The port in which the message has been received
 * @param answer_port The port to which an error (or OK) message will be sent. If it's 0, no message will be sent
 * @param connection The client's connection
 * @param user_data Extra parameter used in callbacks
 *
 * @return 0 on success; -1 on error
 *
 */

uint32_t clients_define_windows(uint32_t port, uint32_t answer_port, struct minibus_connection *connection, void *user_data) {

	uint32_t tmp_nwindows,loop,x,y,w,h;
	uint8_t old_window,new_window,visible;
	int32_t error;
	struct _window *tmp_windows,*old_windows;
	struct minibus_data_element *data;
	struct minibus_list_element *element;
	struct _client_data *client_element;

	if (!other_client_is_wm(connection)) {
		other_send_error(connection,answer_port,TTOS_ERROR_RESTRICTED_COMMAND);
		return -1;
	}

	data=minibus_read_params(connection,"l",&error);

	if(!data) {
		switch(error) {
		case -1:
			other_send_error(connection,answer_port,TTOS_ERROR_NO_MEMORY);
			return -1;
		break;
		case -2:
			other_send_error(connection,answer_port,TTOS_ERROR_CANT_READ_DATA);
			return -1;
		break;
		case -3:
			other_send_error(connection,answer_port,TTOS_ERROR_INCORRECT_FORMAT);
			return -1;
		break;
		}
	}

	tmp_nwindows=data[0].value32;
	free(data);

	tmp_windows=(struct _window *)malloc(tmp_nwindows*sizeof(struct _window));
	if (!tmp_windows) {
		other_send_error(connection,answer_port,TTOS_ERROR_NO_MEMORY);
		return -1;
	}

	for(loop=0;loop<tmp_nwindows;loop++) {

		/* Read four shorts and one byte */

		data=minibus_read_params(connection,"ssssb",&error);

		if(!data) {
			free(tmp_windows);
			switch(error) {
			case -1:
				other_send_error(connection,answer_port,TTOS_ERROR_NO_MEMORY);
				return -1;
			break;
			case -2:
				other_send_error(connection,answer_port,TTOS_ERROR_CANT_READ_DATA);
				return -1;
			break;
			case -3:
				other_send_error(connection,answer_port,TTOS_ERROR_INCORRECT_FORMAT);
				return -1;
			break;
			}
		}

		x=data[0].value16;
		y=data[1].value16;
		w=data[2].value16;
		h=data[3].value16;
		visible=data[4].value8;
		free(data);

		if ((w==0)||(h==0)||(base->w<(x+w-1))||(base->h<(y+h-1))) {
			other_send_error(connection,answer_port,TTOS_ERROR_INVALID_WINDOW_SIZE);
			free(tmp_windows);
			return -1;
		}

		tmp_windows[loop].ox=x;
		tmp_windows[loop].oy=y;
		tmp_windows[loop].w=w;
		tmp_windows[loop].h=h;
		tmp_windows[loop].mx=x+w-1;
		tmp_windows[loop].my=y+h-1;
		tmp_windows[loop].visible=visible;
		if (loop<n_windows) {
			tmp_windows[loop].fg_client=windows[loop].fg_client;
		} else {
			tmp_windows[loop].fg_client=NULL;
		}
	}

	old_windows=windows;
	windows=tmp_windows;
	n_windows=tmp_nwindows;

	for(element=clients.list;element!=NULL;element=element->next) {
		client_element=element->data;
		old_window=client_element->window;
		if (old_window<n_windows)
			new_window=old_window;
		else
			new_window=0;
		client_element->window=new_window;

		// If the window associated to this client has changed in size or position, send the new values
		if ((old_windows[old_window].ox!=windows[new_window].ox) ||
			(old_windows[old_window].oy!=windows[new_window].oy) ||
			(old_windows[old_window].w !=windows[new_window].w) ||
			(old_windows[old_window].h !=windows[new_window].h)) {
				clients_send_resolution(client_element->conn);
		}
	}

	free(old_windows);
	other_send_error(connection,answer_port,TTOS_ERROR_NO_ERROR);
	return 0;
}


/**
 * Sets the visible status of a window
 *
 * The first element, a BYTE integer, must be the window number to be set visible or invisible.
 * Following it will come another BYTE integer (0-> window is not visible, 1-> window is visible).
 *
 * If there's an error, it will send an error message; if not, it will send an OK message
 *
 * @param port The port in which the message has been received
 * @param answer_port The port to which an error (or OK) message will be sent. If it's 0, no message will be sent
 * @param connection The client's connection
 * @param user_data Extra parameter used in callbacks
 *
 * @return 0 on success; -1 on error
 *
 */

uint32_t clients_window_set_visible(uint32_t port, uint32_t answer_port, struct minibus_connection *connection, void *user_data) {

	uint8_t window,visible;
	int32_t error;
	struct minibus_data_element *data;

	if (!other_client_is_wm(connection)) {
		other_send_error(connection,answer_port,TTOS_ERROR_RESTRICTED_COMMAND);
		return -1;
	}

	data=minibus_read_params(connection,"bb",&error);

	if(!data) {
		switch(error) {
		case -1:
			other_send_error(connection,answer_port,TTOS_ERROR_NO_MEMORY);
			return -1;
		break;
		case -2:
			other_send_error(connection,answer_port,TTOS_ERROR_CANT_READ_DATA);
			return -1;
		break;
		case -3:
			other_send_error(connection,answer_port,TTOS_ERROR_INCORRECT_FORMAT);
			return -1;
		break;
		}
	}

	window=data[0].value8;
	visible=data[1].value8;
	free(data);

	if (window>n_windows) {
		other_send_error(connection,answer_port,TTOS_ERROR_WINDOW_OUT_OF_RANGE);
		return -1;
	}
	windows[window].visible=visible;
	other_send_error(connection,answer_port,TTOS_ERROR_NO_ERROR);
	return 0;
}


/**
 * Sets the current window for an specified client. It receives the client's UID as first message parameter (long) and
 * the new window as second parameter (byte).
 *
 * If there's an error, it will send an error message; if not, it will send an OK message
 *
 * @param port The port in which the message has been received
 * @param answer_port The port to which an error (or OK) message will be sent. If it's 0, no message will be sent
 * @param connection The client's connection
 * @param user_data Extra parameter used in callbacks
 *
 * @return 0 on success; -1 on error
 *
 */

uint32_t clients_set_client_window(uint32_t port, uint32_t answer_port, struct minibus_connection *connection, void *user_data) {

	uint32_t client_uid;
	uint8_t new_window,old_window;
	int32_t error;
	struct _client_data *client_element;
	struct minibus_data_element *data;

	if (!other_client_is_wm(connection)) {
		other_send_error(connection,answer_port,TTOS_ERROR_RESTRICTED_COMMAND);
		return -1;
	}

	/* Read a long and a byte */
	data=minibus_read_params(connection,"lb",&error);

	if(!data) {
		switch(error) {
		case -1:
			other_send_error(connection,answer_port,TTOS_ERROR_NO_MEMORY);
			return -1;
		break;
		case -2:
			other_send_error(connection,answer_port,TTOS_ERROR_CANT_READ_DATA);
			return -1;
		break;
		case -3:
			other_send_error(connection,answer_port,TTOS_ERROR_INCORRECT_FORMAT);
			return -1;
		break;
		}
	}

	client_uid=data[0].value32;
	new_window=data[1].value8;
	free(data);

	if (new_window>=n_windows) {
		other_send_error(connection,answer_port,TTOS_ERROR_WINDOW_OUT_OF_RANGE);
		return -1;
	}

	client_element=other_get_client_by_uid(client_uid);

	if (!client_element) {
		other_send_error(connection,answer_port,TTOS_ERROR_INVALID_UID);
		return -1;
	}

	old_window=client_element->window;
	client_element->window=new_window;
	windows[new_window].fg_client=client_element;
	if (windows[old_window].fg_client==client_element) {
		windows[old_window].fg_client=NULL;
		other_send_error(connection,answer_port,TTOS_ERROR_OLD_WINDOW_HAS_NO_FG);
		return 0;
	}
	other_send_error(connection,answer_port,TTOS_ERROR_NO_ERROR);
	return 0;
}

/**
 * Moves a client to FG. This function is called from the clients_move_to_fg callback
 *
 * @param client_uid The client's UID
 *
 * @return False on error, True on success
 */

bool clients_do_move_to_fg(uint32_t client_uid) {

	struct _client_data *client_element,*old_client;

	client_element=other_get_client_by_uid(client_uid);
	if (!client_element) {
		return false;
	}

	/* Inform to the old foreground client that now it is in background.
	 * Only if the old client is different from the client to move to FG */

	old_client=windows[client_element->window].fg_client;
	if ((old_client)&&(old_client->uid!=client_uid)) {
		minibus_start_message(write_conn,TTOS_SYSTEM_PORT_2,TTOS_NO_ANSWER);
		minibus_write_long(write_conn,TTOS_SR_MOVED_TO_BG);
		minibus_server_send(write_conn,old_client->conn,false);
	}
	windows[client_element->window].fg_client=client_element;

	/* Inform to the new foreground client that now it is in foreground */
	minibus_start_message(write_conn,TTOS_SYSTEM_PORT_2,TTOS_NO_ANSWER);
	minibus_write_long(write_conn,TTOS_SR_MOVED_TO_FG);
	minibus_server_send(write_conn,client_element->conn,false);

	return true;
}

/**
 * Moves a client to foreground in its window. It receives the client's UID as first message parameter (long).
 *
 * If there's an error, it will send an error message; if not, it will send an OK message
 *
 * @param port The port in which the message has been received
 * @param answer_port The port to which an error (or OK) message will be sent. If it's 0, no message will be sent
 * @param connection The client's connection
 * @param user_data Extra parameter used in callbacks
 *
 * @return 0 on success; -1 on error
 *
 */

uint32_t clients_move_to_fg(uint32_t port, uint32_t answer_port, struct minibus_connection *connection, void *user_data) {

	uint32_t client_uid;
	int32_t error;
	struct minibus_data_element *data;

	if (!other_client_is_wm(connection)) {
		other_send_error(connection,answer_port,TTOS_ERROR_RESTRICTED_COMMAND);
		return -1;
	}

	data=minibus_read_params(connection,"l",&error);

	if(!data) {
		switch(error) {
		case -1:
			other_send_error(connection,answer_port,TTOS_ERROR_NO_MEMORY);
			return -1;
		break;
		case -2:
			other_send_error(connection,answer_port,TTOS_ERROR_CANT_READ_DATA);
			return -1;
		break;
		case -3:
			other_send_error(connection,answer_port,TTOS_ERROR_INCORRECT_FORMAT);
			return -1;
		break;
		}
	}

	client_uid=data[0].value32;
	free(data);

	if (false==clients_do_move_to_fg(client_uid)) {
		other_send_error(connection,answer_port,TTOS_ERROR_INVALID_UID);
		return -1;
	}

	other_send_error(connection,answer_port,TTOS_ERROR_NO_ERROR);
	return 0;
}
