/*
 * Copyright 2010 (C) Raster Software Vigo (Sergio Costas)
 * This file is part of MINIBUS
 *
 * MINIBUS is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * MINIBUS is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

/*
 * queue.c
 *
 *  Created on: 19/08/2010
 *      Author: raster
 */

#include <unistd.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <sys/select.h>
#include <poll.h>

#include <stdio.h>

#include "minibus.h"

void minibus_close_connection(void *object);
int32_t minibus_receive_part(struct minibus_connection *object);

int32_t minibus_send_part(struct minibus_connection *object, uint8_t *data, uint32_t size);

/**
 * Returns the first connection of a queue. Useful for clients, to get its only connection; servers should rely in loops
 *
 * @param object The minibus_master object from which get the client
 *
 * @return The first connection, or NULL in case of error
 *
 */

struct minibus_connection *minibus_get_first_connection(struct minibus_master *object) {

	if (object==NULL) {
		return (NULL);
	}

	if (object->conections.list==NULL) {
		return (NULL);
	}

	return (object->conections.list->data);

}

struct minibus_connection *minibus_init_conn(struct minibus_master *queue,bool client) {

	struct minibus_connection *connection;

	connection=(struct minibus_connection *)malloc(sizeof(struct minibus_connection));
	if (connection==NULL) {
		return NULL;
	}

	connection->master_queue=queue;
	connection->id_socket=0;
	connection->s_size=0;
	connection->s_pointer=2;
	connection->s_mode=MINIBUS_CONNECTION_IDLE;
	if (client) {
		connection->send_message=malloc(65537);
	} else {
		connection->send_message=NULL;
	}

	connection->r_size=0;
	connection->r_pointer=0;
	connection->r_readed=0;
	connection->r_total=0;
	connection->received_message=NULL;
	connection->r_mode=MINIBUS_CONNECTION_IDLE;

	connection->data=NULL;

	connection->polling=NULL;
	connection->send_remaining=0;
	connection->sent=0;
	connection->type=MINIBUS_CLIENT;
	connection->isclosing=0;
	connection->listener_callback=NULL;
	minibus_list_init(&connection->send_buffer,minibus_destroy_message_element,NULL);
	minibus_list_init(&connection->receive_buffer,minibus_destroy_message_element,NULL);

	return(connection);
}


/**
 * Returns an unique identifier for a connection (currently, it's the socket descriptor)
 *
 * @param object The connection
 *
 * @return An unique identifier for that connection, or -1 in case of failure
 *
 */

int32_t minibus_get_uid(struct minibus_connection *object) {

	return (object->id_socket);

}


/**
 * Starts a new message with the two mandatory parameters (destination port and answer port), allowing
 * to reduce the probability of bugs and mistakes
 *
 * @param connection The connection over which to send the message
 * @param port The destination port
 * @param answer_port The answer port (0 if no answer allowed or needed)
 *
 * @return True on success, False on error
 */

inline bool minibus_start_message(struct minibus_connection *connection, uint32_t port, uint32_t answer_port) {

	if (!connection)
		return false;

	minibus_send_reset(connection);
	minibus_write_long(connection,port);
	minibus_write_long(connection,answer_port);

	return true;
}


/**
 * Adds a listener for a file descriptor, allowing to respond from the main loop to events in it.
 *
 * @param object The queue to which add the listener
 * @param fd The file descriptor to monitor
 * @param events The list of events (like in poll() function) to check
 * @param listener_callback the callback function to be called
 * @param data Private data to send to the callback function
 *
 * @return 0 if success, 1 if failure
 */

int32_t minibus_add_listener(struct minibus_master *object, int fd, short events, void (*listener_callback) (int fd, short revents, void *data), void *data) {


	struct minibus_connection *conn;
	int serversock;

	if ((object==NULL) || (listener_callback==NULL)) {
		return -1;
	}

	conn=minibus_init_conn(object,false);
	if (conn==NULL) {
		return -1;
	}

	conn->type=MINIBUS_LISTENER;
	conn->id_socket=fd;
	conn->listener_callback=listener_callback;
	conn->data=data;
	conn->events=events;
	minibus_list_add(&object->conections,conn);

	if (object->polling) {
		if (object->server)
			serversock=object->id_socket;
		else
			serversock=0;

		free(object->polling);
		object->polling=minibus_fill_polling(object,serversock);
		if (object->polling==NULL) {
			return -1;
		}
	}
	return 0;

}


/**
 * Removes a listener for a file descriptor.
 *
 * @param object The queue from which remove the listener
 * @param fd The file descriptor to remove
 *
 * @return 0 if success, 1 if failure
 */

int32_t minibus_remove_listener(struct minibus_master *object, int fd) {

	struct minibus_connection *conn=NULL;
	struct minibus_list_element *element;

	for(element=object->conections.list;element!=NULL;element=element->next) {
		conn=element->data;
		if (conn->type!=MINIBUS_LISTENER) {
			continue;
		}
		if (conn->id_socket==fd) {
			break;
		}
	}

	if (element==NULL) {
		return -1;
	}

	conn=element->data;

	if (conn==NULL) {
		return -1;
	}

	if ((conn->id_socket!=fd)||(conn->type!=MINIBUS_LISTENER)) {
		return -1;
	}

	minibus_list_remove(element,true);

	return 0;
}

void minibus_free_callback(void *element) {

	if (!element)
		return;

	free(element);
}


/**
 * This is the default callback when the connection to the main system is lost. It just exit.
 *
 * @param connection Not needed
 * @param data Not needed
 * @return
 */

uint32_t minibus_default_closed_connection_callback(struct minibus_connection *connection,void *data) {

	exit(-1);

}

/**
 * Initializates a new queue
 *
 * @param minibus_path Path and name for the UNIX socket to use
 * @param server 0 means that this is a client connecting to an existing queue;
 * 				 1 means that this is a server creating a new queue
 *
 * @return A queue object, or NULL if there's an error
 */

struct minibus_master *minibus_init(char *minibus_path, bool server) {

	int mainsocket;
	struct sockaddr_un local;
	struct minibus_master *object;
	struct minibus_connection *connection=NULL;

	object=(struct minibus_master *)malloc(sizeof(struct minibus_master));
	if (object==NULL) {
		printf("Can't do malloc for init queue\n");
		return (NULL);
	}

	minibus_list_init(&object->conections,minibus_close_connection,NULL);
	minibus_list_init(&object->callbacks,minibus_free_callback,NULL);
	object->loop=true;
	object->forced_exit=false;
	object->default_callback=NULL;
	object->new_connection_callback=NULL;
	object->closed_connection_callback=minibus_default_closed_connection_callback;
	object->timeout_callback=NULL;
	object->new_data=NULL;
	object->closed_data=NULL;
	object->default_data=NULL;
	object->timeout_data=NULL;
	object->polling=NULL;
	object->pollsize_total=0;
	object->pollsize_connections=0;
	object->timeout_mode=2;

	connection=minibus_init_conn(object,true);
	if (connection==NULL) {
		free(object);
		printf("Can't do malloc for init connection\n");
		return (NULL);
	}

	if (server) {
		object->server=1;
		object->send_connection=connection;
		unlink(minibus_path);
	} else {
		object->server=0;
		object->send_connection=NULL;
		minibus_list_add(&object->conections,connection);
	}

	mainsocket=socket(PF_UNIX,SOCK_STREAM,0);
	if(mainsocket==-1) {
		printf("Can't open socket for init queue\n");
		free(object);
		if (connection)
			free(connection);
		return (NULL);
	}

	local.sun_family=AF_UNIX;
	strcpy(local.sun_path,minibus_path);
	if (server) {
		if (-1==bind(mainsocket,(struct sockaddr *)&local,sizeof(struct sockaddr_un))) {
			close(mainsocket);
			free(object);
			if (connection)
				free(connection);
			printf("Can't do bind for init queue\n");
			return (NULL);
		}
		chmod(minibus_path,S_IROTH|S_IWOTH|S_IRGRP|S_IWGRP|S_IRUSR|S_IWUSR);
		listen(mainsocket,20);
	} else {
		if (-1==connect(mainsocket,(struct sockaddr *)&local,sizeof(struct sockaddr_un))) {
			close(mainsocket);
			free(object);
			if (connection)
				free(connection);
			printf("Can't do connect for init queue\n");
			return (NULL);
		}
	}

	if (server) {
		object->id_socket=mainsocket;
	} else {
		connection->id_socket=mainsocket;
		object->id_socket=0;
	}

	return (object);

}

/**
 * Closes a queue and frees all the structures used in it
 *
 * @param object The queue to release
 *
 * @return 0 if all went fine; -1 if there was an error
 */

int32_t minibus_release(struct minibus_master *object) {

	if (object==NULL) {
		return -1;
	}

	minibus_list_free(&object->conections);
	minibus_list_free(&object->callbacks);
	if (object->send_connection) {
		minibus_close_connection(object->send_connection);
	}
	minibus_destroy_polling(object);
	free(object);
	return 0;

}

/**
 * Waits until a client connects to a queue
 *
 * @param object A queue object created as server
 * @return 0 if all went fine, -1 if there was an error
 */

int32_t minibus_accept(struct minibus_master *object) {

	struct sockaddr_un remote;
	socklen_t size;
	int retval;
	struct minibus_connection *connection;

	if (!object) {
		return -1;
	}

	if (object->id_socket==0) {
		return 0;
	}

	size = sizeof(struct sockaddr_un);
	retval=accept(object->id_socket,(struct sockaddr *)&remote,&size);
	if (retval>=0) {
		connection=minibus_init_conn(object,false);
		if (connection==NULL) {
			printf("Can't accept a connection\n");
			return -1;
		}
		connection->id_socket=retval;
		minibus_list_add(&object->conections,connection);
		retval=0;
		if (object->new_connection_callback) {
			object->new_connection_callback(connection,object->new_data);
		}
	}

	return (retval);
}

/**
 * Closes an specific connection and releases all the memory used by it
 *
 * @param object A connection object
 */

void minibus_close_connection(void *element) {

	struct minibus_connection *object=element;

	if (object==NULL) {
		return;
	}

	if (object->type==MINIBUS_CLIENT) {
		close(object->id_socket);
		minibus_list_free(&object->send_buffer);
		minibus_list_free(&object->receive_buffer);
	}
	if (object->received_message) {
		free(object->received_message);
	}
	if (object->send_message) {
		free(object->send_message);
	}
	free(object);
}

/**
 * Resets connection to start again to create a new message to send
 *
 * @param object A connection object
 */

void minibus_send_reset(struct minibus_connection *object) {

	if (object==NULL) {
		return;
	}

	object->s_pointer=2;
	object->s_size=0;
}

int32_t minibus_receive_part(struct minibus_connection *object) {

	int size;
	uint8_t tmpbuf[2];

	// 0: message read; process it
	// 1: message half-read; continue reading
	// 2: error during reading; aborted

	if (object->r_mode==MINIBUS_CONNECTION_IDLE) { // we are starting a new receiving process
		object->r_mode=MINIBUS_CONNECTION_RECEIVING;
		object->r_total=2;
		object->r_readed=0;
		do {
			size=read(object->id_socket,tmpbuf+object->r_readed,2-object->r_readed);
			if (size<=0) {
				object->r_size=0;
				object->r_mode=MINIBUS_CONNECTION_IDLE; // no message; allow to write
				return 2;
			}
			object->r_readed+=size;
		} while(object->r_readed<2); // wait until we have the message size

		if (object->received_message) {
			free(object->received_message);
		}
		object->r_pointer=2;
		object->r_size=((((uint16_t)(tmpbuf[0]))<<8)&0xFF00)|(((uint16_t)tmpbuf[1])&0x00FF);
		object->r_total=2+object->r_size;
		object->received_message=malloc(2+object->r_size);
		object->received_message[0]=tmpbuf[0];
		object->received_message[1]=tmpbuf[1];
	}

	size=read(object->id_socket,object->received_message+object->r_readed,object->r_total-object->r_readed);
	if (size<=0) {
		object->r_size=0;
		object->r_mode=MINIBUS_CONNECTION_IDLE; // no message; allow to write
		return 2;
	}
	object->r_readed+=size;

	if (object->r_readed>=object->r_total) {
		object->r_mode=MINIBUS_CONNECTION_READ; // message read, we can proceed
		object->r_pointer=2;
		return 0;
	} else {
		return 1;
	}
}

/**
 * Waits until a message arrives. When it returns, the message can be processed.
 *
 * @param object A connection object
 * @param msecs Timeout in milliseconds (0 for wait infinite time)
 *
 * @return 0: message fully read; 1: timeout; -1: the object is invalid; -2: there was an error while receiving
 */

int32_t minibus_receive(struct minibus_connection *object, int32_t msecs) {

	int32_t retval;
	struct pollfd fdl;

	if (object==NULL) {
		return -1;
	}

	if ((object->id_socket==0)||(object->isclosing)) {
		return (-1);
	}

	if (msecs>0) {
		fdl.fd=object->id_socket;
		fdl.events=POLLIN;
		if (1!=poll(&fdl,1,(int)msecs)) {
			return 1;
		}
	}

	while(1==(retval=minibus_receive_part(object)));

	if (retval) {
		return (-2);
	} else {
		return (0);
	}
}


/**
 * Properly destroys the polling object, setting all the polling pointers to NULL
 * @param object The master queue whose polling object destroy
 */

void minibus_destroy_polling(struct minibus_master *object) {

	struct minibus_list_element *element;
	struct minibus_connection *conn;

	if (!object) {
		return;
	}

	if (object->polling) {
		free(object->polling);
		object->polling=NULL;
	}

	for(element=object->conections.list;element;element=element->next) {
		conn=element->data;
		conn->polling=NULL;
	}
}


/**
 * Fills the pollfd structure with the socket descriptors currently open
 */

struct pollfd *minibus_fill_polling(struct minibus_master *object, int serversock) {

	struct pollfd *polling;
	struct minibus_list_element *element;
	struct minibus_connection *conn;
	uint32_t counter,counter2,addc;

	if (serversock) {
		addc=1;
	} else {
		addc=0;
	}

	for(element=object->conections.list,counter=addc,counter2=addc;element!=NULL;element=element->next) {
		counter++;
		conn=element->data;
		if (conn->type==MINIBUS_CLIENT) {
			counter2++;
		}
	}
	object->pollsize_total=counter;
	object->pollsize_connections=counter2;

	polling=(struct pollfd *)malloc(counter*sizeof(struct pollfd)); // one more to count the server socket, if needed
	if (polling==NULL) {
		return (NULL);
	}

	if (addc) {
		polling[0].fd=serversock;
		polling[0].events=POLLIN;
		polling[0].revents=0;
	}

	for(element=object->conections.list,counter=addc;element!=NULL;element=element->next) {
		conn=(struct minibus_connection *)element->data;
		if (conn->type!=MINIBUS_CLIENT) {
			continue;
		}
		polling[counter].fd=conn->id_socket;
		if (conn->send_buffer.list!=NULL) {
			polling[counter].events=POLLIN|POLLOUT|POLLHUP;
		} else {
			polling[counter].events=POLLIN|POLLHUP;
		}
		polling[counter].revents=0;
		conn->polling=&(polling[counter]);
		counter++;
	}

	for(element=object->conections.list;element!=NULL;element=element->next) {
		conn=(struct minibus_connection *)element->data;
		if (conn->type!=MINIBUS_LISTENER) {
			continue;
		}
		polling[counter].fd=conn->id_socket;
		polling[counter].events=conn->events;
		polling[counter].revents=0;
		conn->polling=&(polling[counter]);
		counter++;
	}
	return (polling);
}


/**
 * Reads a complete message from a socket
 *
 * @param conn Connection object from which to read the message
 * @param object The queue
 *
 * @return 1 on success; -1 if there's an error; 0 if has not read a full message
 */

int32_t minibus_do_read_from_loop(struct minibus_connection *conn,struct minibus_master *object,struct minibus_wait_data *retval) {

	struct minibus_data_element data;

	if (0!=minibus_receive_part(conn)) {
		return 0;
	}

	minibus_read_data(conn,&data);
	if (data.type!=MINIBUS_TYPE_INT32) {
		return -1;
	}
	retval->port=data.value32;
	retval->conn=conn;
	return 1;
}

/**
 * Calls the corresponding callback for the message type received, acording to wdata
 *
 * @param object The queue
 * @param wdata The message type and the connection where the message was received
 *
 * @return -1 if a malformed message arrived; 0 if all went ok
 */

int32_t minibus_do_callback(struct minibus_master *object, struct minibus_wait_data *wdata) {

	struct minibus_callback *cb;
	struct minibus_list_element *element;
	struct minibus_data_element data;
	uint32_t msgtype,answer_port;
	bool has_callback;

	msgtype=wdata->port;

	minibus_read_data(wdata->conn,&data);
	if (data.type!=MINIBUS_TYPE_INT32) {
		return -1;
	}
	answer_port=data.value32;

	has_callback=false;
	for(element=object->callbacks.list;element!=NULL;element=element->next) {
		cb=element->data;
		if(cb->port==msgtype) {
			cb->function(msgtype,answer_port,wdata->conn,cb->data);
			has_callback=true;
			break;
		}
	}

	if ((!has_callback)&&(object->default_callback)) {
		object->default_callback(msgtype,answer_port,wdata->conn,object->default_data);
	}
	wdata->conn->r_mode=MINIBUS_CONNECTION_IDLE;

	return 0;
}


int32_t minibus_do_write_from_loop(struct minibus_connection *conn,struct minibus_master *object) {

	struct minibus_message_element *element;

	if (conn->send_buffer.list) {
		element=(struct minibus_message_element *)conn->send_buffer.list->data;
		if (1!=minibus_send_part(conn,element->data,element->size)) {
			minibus_list_remove(conn->send_buffer.list,true);
		}
	}

	return 0;
}


/**
 * Returns the private data set by the user. This allows to do struct minibus_connection a fully opaque pointer
 *
 * @param object A connection object
 *
 * @return The pointer set by the user previously
 */

void *minibus_get_connection_data(struct minibus_connection *object) {

	if (!object) {
		return (NULL);
	}

	return (object->data);

}


/**
 * Sets the private data. This allows to do struct minibus_connection a fully opaque pointer
 *
 * @param object A connection object
 * @param data Data pointer
 *
 */

void minibus_set_connection_data(struct minibus_connection *object,void *data) {

	if (object)
		object->data=data;

}

/**
 * Adds a callback for the "port" (or "message type") specified. This callback will be called each time a message
 * is received (in the main loopback) with that destination set (contained in the first four bytes of the message, as
 * a long int). Only one callback per port is allowed; assigning several times a callback result in only the last
 * function being called.
 *
 * The callback function receives the port number (so the same callback can manage several ports), the connection object
 * (allowing it to read the rest of the message, and send back and receive more messages), and a pointer to private data
 * set by the user in this call.
 *
 * @param object The queue object where to add the callback
 * @param port The port number
 * @param function Pointer to the callback function
 * @param data Private data to be sent to the callback
 *
 * @return Returns 0 on success, or a negative value if the callback can't be added
 */

int32_t minibus_add_callback(struct minibus_master *object, uint32_t port, uint32_t (*function)(uint32_t port, uint32_t answer_port, struct minibus_connection *queue, void *data),void *data) {

	struct minibus_callback *cb;
	bool found;
	struct minibus_list_element *element;
	int serversock;

	if (function==NULL) {
		return -2;
	}

	found=false;
	for(element=object->callbacks.list,found=0;element!=NULL;element=element->next) {
		cb=element->data;
		if(cb->port==port) {
			found=true;
			break;
		}
	}

	if (!found) {
		cb=(struct minibus_callback *)malloc(sizeof(struct minibus_callback));
		if (cb==NULL) {
			return -1;
		}
	}

	cb->data=data;
	cb->port=port;
	cb->function=function;

	if (!found) {
		minibus_list_add(&object->callbacks,cb);
	}

	if (object->polling) {
		if (object->server)
			serversock=object->id_socket;
		else
			serversock=0;

		free(object->polling);
		object->polling=minibus_fill_polling(object,serversock);
		if (object->polling==NULL) {
			return -1;
		}
	}

	return 0;
}


/**
 * Removes a callback from a queue.
 *
 * @param object The queue from where to remove the callback
 * @param port The port to remove
 *
 * @return 0 on success; -1 on failure
 */

int32_t minibus_remove_callback(struct minibus_master *object, uint32_t port) {

	struct minibus_callback *cb;
	struct minibus_list_element *element;

	for(element=object->callbacks.list;element!=NULL;element=element->next) {
		cb=element->data;
		if(cb->port==port) {
			minibus_list_remove(element,true);
			return 0;
		}
	}
	return -1;
}


/**
 * Sets a callback function that will be called when arrives a message to a port without an assigned callback. Setting the
 * callback to NULL removes it.
 *
 * The callback receives the port number (so the same callback can manage several ports), the connection object (allowing
 * it to read the rest of the message, and send back and receive more messages), and a pointer to private data set by the
 * user in this call.
 *
 * @param object The queue object where to add the callback
 * @param function Pointer to the callback function
 * @param data Private data to be sent to the callback
 */

void minibus_set_default_callback(struct minibus_master *object, uint32_t (*function)(uint32_t port, uint32_t answer_port, struct minibus_connection *queue, void *data),void *data) {

	object->default_callback=function;
	object->default_data=data;

}


/**
 * Sets a callback function that will be called after a new connection is created.
 *
 *  The callback receives the new connection object (allowing it to set private data and more) and a pointer to private
 *  data set by the user in this call.
 *
 * @param object The queue object where to add the callback
 * @param new_connection_callback Pointer to the callback function
 * @param data Private data to be sent to the callback
 */

void minibus_set_new_connection_callback(struct minibus_master *object, uint32_t (*new_connection_callback)(struct minibus_connection *connection,void *data),void *data) {

	object->new_connection_callback=new_connection_callback;
	object->new_data=data;

}


/**
 * Sets a callback function that will be called before a connection is destroyed after having been closed by the other side.
 *
 *  The callback receives the connection object (allowing it to destroy private data and more) and a pointer to private
 *  data set by the user in this call.
 *
 * @param object The queue object where to add the callback
 * @param closed_connection_callback Pointer to the callback function
 * @param data Private data to be sent to the callback
 */

void minibus_set_closed_connection_callback(struct minibus_master *object, uint32_t (*closed_connection_callback)(struct minibus_connection *connection,void *data),void *data) {

	object->closed_connection_callback=closed_connection_callback;
	object->closed_data=data;

}


/**
 * Sets a callback function that will be called each time a receiving timeout occurs, or a message is received.
 * When a message is received, first its callback will be called (or the default callback), and only then this callback
 * will be called, after which the system will wait for a new message. This callback is also called when a new connection
 * has been made or when a connection has been closed, in order to set the new timeout if needed. It's also called before
 * the first time the loop waits for a message, in order to set the timeout in that case.
 *
 * The callback receives three params: the queue object currently used, the timeout type (0: no timeout, a message has
 * arrived; 1: timeout, no new messages arrived since last call; 2: no timeout, first call), and a pointer with private
 * data. It must return the milliseconds for next timeout (or -1 to wait forever until a new message)
 *
 * @param object The queue object where to set the callback
 * @param timeout_callback The callback funcion
 * @param data Private data to be sent to the callback
 *
 */

void minibus_set_timeout_callback(struct minibus_master *object, int32_t (*timeout_callback)(struct minibus_master *object, uint32_t timeout,void *data),void *data) {

	object->timeout_callback=timeout_callback;
	object->timeout_data=data;

}


/**
 * Closes a connection based on its socket, destroying all its associated objects and calling the destroy callback.
 *
 * @param conn The connection object to close
 * @param object The queue object to where the connection belongs
 *
 * @return 0 on success, -1 on error (if there's no connection with that socket)
 */

int32_t minibus_do_close_conn(struct minibus_connection *conn,struct minibus_master *object) {

	struct minibus_list_element *element;

	if ((!object)||(!conn)) {
		return -1;
	}

	for(element=object->conections.list;element!=NULL;element=element->next) {
		if (element->data==conn) {
			break;
		}
	}
	if (element==NULL) {
		return -1;
	}

	if (object->closed_connection_callback) {
		object->closed_connection_callback(conn,object->closed_data);
	}

	minibus_list_remove(element,true);

	return 0;
}


/**
 * Exits the main loop in the queue
 *
 * @param object The queue object
 *
 * @return 0 on success, -1 on error
 */

int32_t minibus_exit_loop(struct minibus_master *object) {

	if (object==NULL) {
		return -1;
	}

	object->loop=false;

	return 0;
}

/**
 * Waits for a new message to arrive
 *
 * @param object The queue where to wait for messages
 * @param serversock The server socket if this is a server queue; 0 if not
 * @param retval An struct where to return the message type and the connection when a message arrives
 *
 * @return 0 if it has to exit the main loop; 1 if a message arrived; 2 if there was a timeout; -1 if there was an error;
 */


int32_t minibus_read_message(struct minibus_master *object,int serversock,struct minibus_wait_data *retval, char wait_for_listeners) {

	struct minibus_list_element *element;
	struct minibus_connection *conn;
	int pollretval;
	short int tmp_revents;
	int32_t timeout;

	while(1) {

		if (!object->loop)
			return 0; // In case the callback itself forces the exit of the loop

		if (object->timeout_callback) {
			timeout=object->timeout_callback(object,object->timeout_mode,object->timeout_data);
		} else {
			timeout=-1;
		}

		pollretval=poll(object->polling,wait_for_listeners==true ? object->pollsize_total : object->pollsize_connections,timeout);

		// Timeout_mode will be 1 if there was a real timeout, or 0 if there was a message
		object->timeout_mode = (pollretval==0) ? 1 : 0;

		if (pollretval>0) {
			if (object->server) {
				if (object->polling[0].revents) { // we are a server and there's a new connection
					minibus_accept(object);
					free(object->polling);
					object->polling=minibus_fill_polling(object,serversock); // refresh the list of connections to poll
					if (!object->polling) {
						return -1;
					}
					continue;
				}
			}

			for(element=object->conections.list;element!=NULL;element=element->next) {
				conn=(struct minibus_connection *)element->data;
				if (!conn->polling) {
					continue;
				}
				if (conn->type==MINIBUS_CLIENT) {
					if ((conn->polling->revents)&POLLHUP) { // socket closed
						conn->polling->revents=0;
						conn->isclosing=1;
						minibus_do_close_conn(conn,object); // release all the data for this connection
						free(object->polling);
						object->polling=minibus_fill_polling(object,serversock);
						if (object->polling==NULL) {
							return -1;
						}
						break;
					}
					if ((conn->polling->revents)&POLLIN) { // there's data to read
						if (1==minibus_do_read_from_loop(conn,object,retval)) {
							conn->polling->revents=0;
							return 1;
						}
					}

					if ((conn->polling->revents)&POLLOUT) { // there's data to write
						minibus_do_write_from_loop(conn,object);
					}
					conn->polling->revents=0;
					if (conn->send_buffer.list) {
						conn->polling->events=POLLHUP|POLLIN|POLLOUT;
					} else {
						conn->polling->events=POLLHUP|POLLIN;
					}
				} else {
					if (conn->polling->revents) {
						tmp_revents=conn->polling->revents;
						conn->polling->revents=0;
						conn->polling->events=conn->events;
						conn->listener_callback(conn->id_socket,tmp_revents,conn->data);
					}
				}
			}
		} else {
			if (retval==0) {
				return 2; // timeout
			}
		}
	}
}



void minibus_run_pending_callbacks(struct minibus_master *object, struct minibus_connection *conn) {

	uint32_t size,pointer;
	uint8_t *c_message;
	struct minibus_wait_data port_conn;
	struct minibus_list_element *element;
	struct minibus_message_element *message;

	// store the current values
	size=conn->r_size;
	pointer=conn->r_pointer;
	c_message=conn->received_message;

	port_conn.conn=conn;
	while(conn->receive_buffer.list) {
		element=conn->receive_buffer.list;
		message=element->data;
		conn->received_message=message->data;
		conn->r_pointer=message->pointer;
		conn->r_size=message->size;
		port_conn.port=message->event;
		minibus_do_callback(object,&port_conn);
		free(conn->received_message);
		conn->received_message=NULL;
		message->data=NULL; // this way, if conn->received_message has been replaced during the callback, will not fail the
							// removing of the element in the next call (minibus_list_remove) due to double_free.
		minibus_list_remove(element,true);
	}
	conn->r_size=size;
	conn->r_pointer=pointer;
	conn->received_message=c_message;
}

/**
 * Enters in a loop which checks all the current open connections, accepts new ones, releases the closed ones, read the
 * messages received in each and calls the associated callbacks.
 *
 * @param object The queue object
 *
 * @return 0 on success; -1 if there was a memory allocation error
 */

int32_t minibus_run_loop(struct minibus_master *object) {

	int serversock;
	int32_t retval;
	struct minibus_wait_data port_conn;
	struct minibus_connection *conn;
	struct minibus_list_element *element;

	if (object->server)
		serversock=object->id_socket;
	else
		serversock=0;

	if (object->polling) {
		free(object->polling);
	}

	object->polling=minibus_fill_polling(object,serversock);
	if (!object->polling) {
		return -1;
	}

	object->loop=true;

	// run all pending events before entering the main loop
	for(element=object->conections.list;element!=NULL;element=element->next) {
		conn=element->data;
		minibus_run_pending_callbacks(object,conn);
	}

	while(object->loop) {
		retval=minibus_read_message(object,serversock,&port_conn,true);
		if (retval<0) {
			return (retval);
		}
		if (retval==1) { // a message arrived; do its callback
			minibus_do_callback(object,&port_conn);
			minibus_run_pending_callbacks(object,port_conn.conn);
		}
	}

	minibus_destroy_polling(object);
	return 0;
}


/**
 * Fills the two first bytes in a message with the size
 */

void minibus_fill_size(struct minibus_connection *object) {

	uint32_t value=object->s_size;

	// write the message size
	// it adds 2 to SIZE to take in count the SIZE field

	object->send_message[0] = (uint8_t) ((value>>8)&0x0FF);
	object->send_message[1] = (uint8_t) (value&0x0FF);
	object->s_size+=2;

}



/**
 * Creates a minibus_message_element object from the current message's send buffer, ready to be added to a send queue
 *
 * @param object The connection object from which buffer the data will be taken
 *
 * @return A minibus_send_element object with the data
 */

struct minibus_message_element *minibus_create_send_message_element(struct minibus_connection *object) {

	uint32_t size;
	uint8_t *buffer;
	struct minibus_message_element *element;

	size=object->s_size+2;
	buffer=(uint8_t *)malloc(size);
	if (buffer==NULL) {
		return (NULL);
	}
	element=(struct minibus_message_element *)malloc(sizeof(struct minibus_message_element));
	if (element==NULL) {
		free(buffer);
		return (NULL);
	}

	minibus_fill_size(object);
	memcpy(buffer,object->send_message,size);
	element->data=buffer;
	element->size=size;
	element->count=1;

	object->s_size=0;
	object->s_pointer=2;
	object->s_mode=MINIBUS_CONNECTION_IDLE;

	return (element);
}


/**
 * Creates a minibus_message_element object from the current message's receive buffer, ready to be added to a receive queue
 *
 * @param object The connection object from which buffer the data will be taken
 *
 * @return A minibus_send_element object with the data
 */

struct minibus_message_element *minibus_create_receive_message_element(struct minibus_connection *object, uint32_t event) {

	uint32_t size;
	struct minibus_message_element *element;

	size=object->r_size+2;
	element=(struct minibus_message_element *)malloc(sizeof(struct minibus_message_element));
	if (element==NULL) {
		return (NULL);
	}

	element->data=object->received_message;
	object->received_message=NULL;
	element->size=object->r_size;
	element->count=1;
	element->event=event;
	element->pointer=object->r_pointer;

	object->r_size=0;
	object->r_pointer=2;
	object->r_mode=MINIBUS_CONNECTION_IDLE;

	return (element);
}


/**
 * Decrements the reference counter of the send_element passed as parameter, and, if 0, frees it. This allows to
 * send the same message to several connections.
 *
 * @param element A minibus_send_element object
 */

void minibus_destroy_message_element (void *element) {

	struct minibus_message_element *object=element;

	if (object==NULL) {
		return;
	}

	object->count--;
	if (!object->count) {
		if (object->data) {
			free(object->data);
		}
		free(object);
	}
}


/**
 * Sends part
 */

int32_t minibus_send_part(struct minibus_connection *object, uint8_t *data, uint32_t size) {

	int sent;

	// 0: message sent; next
	// 1: message half-sent; continue sending
	// 2: error during writting; aborted

	if (object->sent==0) {
		object->send_remaining=size;
	}

	sent=write(object->id_socket,data+object->sent,object->send_remaining);
	if (sent<0) {
		return -2;
	}

	object->sent+=(uint16_t)sent;
	object->send_remaining-=(uint16_t)sent;

	if (object->send_remaining) {
		return 1;
	} else {
		object->sent=0;
		return 0;
	}
}

/**
 * Sends over a connection the message previously composed and any pending message
 *
 * @param object The connection object
 *
 * @return 0 on success; -1 on error.
 */

int32_t minibus_send(struct minibus_connection *object) {

	return(minibus_server_send(object,object,false));
}

int32_t minibus_server_send(struct minibus_connection *from_object,struct minibus_connection *to_object,bool resend) {

	struct minibus_message_element *element;
	uint8_t *old_pointer;
	uint32_t old_size;

	if ((from_object==NULL)||(to_object==NULL)) {
		return -1;
	}

	if ((to_object->id_socket==0)||(to_object->isclosing)) {
		return -2;
	}

	if (resend) {
		old_size=from_object->s_size;
		from_object->s_size=from_object->r_size;
		old_pointer=from_object->send_message;
		from_object->send_message=from_object->received_message;
	}

	if (from_object->s_size==0) {
		return -3; // no data to send
	}

	element=minibus_create_send_message_element(from_object);

	if (resend) {
		from_object->s_size=old_size;
		from_object->send_message=old_pointer;
	}

	minibus_list_add(&to_object->send_buffer,element);
	// if there's a polling object, then we are inside a main loop, so we just store it ready to be sent ASAP
	if (to_object->polling) {
		to_object->polling->events|=POLLOUT;
	} else {
	// if not, we send everything now
		while(to_object->send_buffer.list) {
			element=(struct minibus_message_element *)to_object->send_buffer.list->data;
			while(1==minibus_send_part(to_object,element->data,element->size));
			minibus_list_remove(to_object->send_buffer.list,true);
		}
	}

	from_object->s_size=0;
	from_object->s_pointer=2;
	from_object->s_mode=MINIBUS_CONNECTION_IDLE;

	return (0);

}
