/*
 * Copyright 2010 (C) Raster Software Vigo (Sergio Costas)
 * This file is part of TTOS
 *
 * TTOS is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * TTOS is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

/*
 * portable.c
 *
 *  Created on: 01/09/2010
 *      Author: raster
 */

#include <sys/mman.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/ioctl.h>

#ifdef TTOS_USE_SDL
#include <pthread.h>
#include <SDL.h>
static SDL_Surface *SDLSCREEN;
#endif

#if defined(TTOS_USE_FB) || defined(TTOS_USE_FAKE_FB)
#include <linux/fb.h>
#endif

#include "portable.h"

static TTOS_Surface screen;
static uint8_t bpp; /* bytes per pixel */
static long int jump,o_line,d_line; /* needed for dump screen */
static uint8_t rotation; /* Current screen rotation */
static uint8_t ts_rotation; /* Current screen rotation */
static int32_t s_width;
static int32_t s_height;


#if defined(TTOS_USE_SDL) || defined(TTOS_USE_FAKE_FB)
static int pipefd[2]; /* pipes for touch screen emulation in SDL */
pthread_t ts_thread; /* thread which reads the mouse under SDL emulation */
volatile char keep_thread;
#endif

#ifdef TTOS_USE_FB
static int fb_descriptor; /* File descriptor for the FrameBuffer */
#endif

static enum TTOS_TOUCH_TYPES ts_status;
static int ts_x,ts_y;

/**
 * Initializates the screen subsystem.
 *
 * @param do_rotation The rotation to apply to the screen (0: no rotation; 1: clockwise; 2: upside down; 3: counter-clockwise)
 *
 * @return A TTOS_Surface ready to be used as buffer to paint in the screen, or NULL if there's an error
 */

TTOS_Surface *ttos_init_screen(uint8_t do_rotation,uint8_t portrait, uint16_t width,uint16_t height) {

#ifdef TTOS_USE_FB
	struct fb_var_screeninfo var;
#endif

	rotation=do_rotation;
	TTOS_Surface *buffer;
	buffer=(TTOS_Surface *)malloc(sizeof(TTOS_Surface));
	if (!buffer) {
		return (NULL);
	}

#ifdef TTOS_USE_FAKE_FB
	buffer->w=SCREEN_WIDTH;
	buffer->h=SCREEN_HEIGHT;
#endif

#ifdef TTOS_USE_SDL

	SDL_Init(SDL_INIT_VIDEO);
	if ((rotation==0)|| (rotation==2)) {
		SDLSCREEN=SDL_SetVideoMode(width, height, 16, SDL_SWSURFACE);
		s_width=width;
		s_height=height;
	} else {
		SDLSCREEN=SDL_SetVideoMode(height, width, 16, SDL_SWSURFACE);
		s_width=height;
		s_height=width;
	}
	buffer->w=width;
	buffer->h=height;
	screen.w=width;
	screen.h=height;
	screen.pixels=SDLSCREEN->pixels;
	bpp=2;
#endif

#ifdef TTOS_USE_FB
	fb_descriptor=open("/dev/fb",O_RDWR);
	if (fb_descriptor<0) {
		exit(1);
	}

	if (ioctl(fb_descriptor, FBIOGET_VSCREENINFO, &var) < 0) {
		exit(2);
	}

	screen.pixels=mmap(NULL,var.xres*var.yres*2,PROT_READ|PROT_WRITE,MAP_SHARED,fb_descriptor,0);
	if (screen.pixels==NULL) {
		exit(2);
	}
	s_width=var.xres;
	s_height=var.yres;
	if (((portrait==0)&&(var.xres<var.yres))||((portrait==1)&&(var.xres>var.yres))) {
		rotation++;
		if (rotation>3) {
			rotation=0;
		}
	}
	if ((rotation==0)|| (rotation==2)) {
		screen.w=var.xres;
		screen.h=var.yres;
		buffer->w=var.xres;
		buffer->h=var.yres;
	} else {
		screen.w=var.yres;
		screen.h=var.xres;
		buffer->w=var.yres;
		buffer->h=var.xres;
	}
	bpp=var.bits_per_pixel / 8;
#endif

	screen.alpha=NULL;

	switch(rotation) {
	case 0:
		o_line=0;
		jump=1;
		d_line=0;
		break;
	case 2:
		o_line=(screen.w*screen.h-1);
		jump=-1;
		d_line=0;
		break;
	case 1:
		o_line=(screen.h-1);
		jump=screen.h;
		d_line=-(((screen.h*screen.w)+1));
		break;
	case 3:
		o_line=((screen.w-1)*screen.h);
		jump=-screen.h;
		d_line=((screen.w*screen.h)+1);
		break;
	}
	buffer->pixels=(uint16_t *)malloc(buffer->w*buffer->h*2);
	buffer->alpha=NULL;

	return(buffer);
}

void ttos_destroy_screen(TTOS_Surface *surface) {

	if (surface->pixels)
		free(surface->pixels);
	if (surface->alpha)
		free(surface->alpha);
	free(surface);

#ifdef TTOS_USE_SDL

	keep_thread=0;
	pthread_join(ts_thread,NULL);
	SDL_QuitSubSystem(SDL_INIT_VIDEO);
	SDL_Quit();
#endif

#ifdef TTOS_USE_FB
	munmap(screen.pixels,2*screen.w*screen.h);
	close(fb_descriptor);
#endif

}

/**
 * Copies a zone from the user screen buffer to the physical screen, converting the colors
 *
 * @param origin_surface Pointer to the surface with the picture to be put in screen. It must have exactly the same size than
 * the screen
 *
 * @return 0 on success; -1 on error
 */


uint32_t ttos_dump_screen(TTOS_Surface *origin_surface) {

	uint32_t lx,ly;
	uint16_t *screen_data,*origin;
	uint32_t w,h;

#if 0
	int fd;
	static int counter=0;
	char fname[256];
	char header[]={0x42,0x4D,
			      0x7A,0x58,0x02,0x00,
			      0,0,
			      0,0,
			      122,0,0,0,

			      108,0,0,0,
			      0x40,0x01,0,0,
			      0xF0,0,0,0,
			      1,0,
			      16,0,
			      3,0,0,0,
			      0x00,0x58,0x02,0x00,
			      19,11,0,0,
			      19,11,0,0,
			      0,0,0,0,
			      0,0,0,0,
			      0x00,0xF8,0,0,
			      0xE0,0x07,0,0,
			      0x1F,0x00,0,0,
			      0x00,0x00,0,0,
			      0x20,0x6E,0x69,0x57,
			      0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
			      0,0,0,0,
			      0,0,0,0,
			      0,0,0,0
	};
	sprintf(fname,"anim/output%06d.bmp",counter);
	counter++;
	fd=open(fname,O_WRONLY|O_CREAT,0666);
	write(fd,header,122);
	write(fd,origin_surface->pixels,320*240*2);
	close(fd);
#endif


#ifdef TTOS_USE_FAKE_FB
	return 0;
#endif

	if (origin_surface==NULL) {
		return -1;
	}

	if ((screen.w!=origin_surface->w)||(screen.h!=origin_surface->h)) {
		return -1;
	}

	w=screen.w;
	h=screen.h;

#if TTOS_USE_SDL
	if ( SDL_MUSTLOCK(SDLSCREEN) ) {
		SDL_LockSurface(SDLSCREEN);
	}
#endif

	origin=origin_surface->pixels;
	screen_data=screen.pixels+o_line;

	if (bpp==2) {
		for(ly=0;ly<h;ly++) {
			for(lx=0;lx<w;lx++) {
				*(screen_data)=*(origin++);
				screen_data+=jump;
			}
			screen_data+=d_line;
		}
	}

#if TTOS_USE_SDL
	if ( SDL_MUSTLOCK(SDLSCREEN) ) {
		SDL_UnlockSurface(SDLSCREEN);
	}
	SDL_Flip(SDLSCREEN);
#endif

	return 0;
}

#ifdef TTOS_USE_SDL

void * ttos_emulate_ts_SDL(void *nothing) {

	SDL_Event event;
	SDL_MouseButtonEvent *bevent;
	SDL_MouseMotionEvent *mevent;
	unsigned char data[5];

	printf("Starting thread\n");

	keep_thread=1;

	while(keep_thread) {

		if (!SDL_WaitEvent(&event)) {
			continue;
		}
		switch (event.type) {
		case(SDL_MOUSEBUTTONDOWN):
		case(SDL_MOUSEBUTTONUP):
			bevent=(SDL_MouseButtonEvent *)(&event);
			data[0]=(unsigned char)(0x00FF & bevent->x);
			data[1]=(unsigned char)(0x00FF & (bevent->x >> 8));
			data[2]=(unsigned char)(0x00FF & bevent->y);
			data[3]=(unsigned char)(0x00FF & (bevent->y >> 8));
			data[4]= (event.type==SDL_MOUSEBUTTONDOWN) ? 1 : 0;
			write(pipefd[1],data,5);
		break;
		case(SDL_MOUSEMOTION):
			mevent=(SDL_MouseMotionEvent *)(&event);
			if (mevent->state&SDL_BUTTON(1)) {
				data[0]=(unsigned char)(0x00FF & mevent->x);
				data[1]=(unsigned char)(0x00FF & (mevent->x >> 8));
				data[2]=(unsigned char)(0x00FF & mevent->y);
				data[3]=(unsigned char)(0x00FF & (mevent->y >> 8));
				data[4]=1;
				write(pipefd[1],data,5);
			}
		break;
		case(SDL_QUIT):
			SDL_Quit();
			exit(0);
		break;
		default:
			continue;
		break;
		}

	}

	return (NULL);
}

#endif

void ttos_destroy_touchscreen(struct tsdev *object) {

#if defined(TTOS_USE_FAKE_FB) || defined(TTOS_USE_SDL)
	close(pipefd[0]);
	close(pipefd[1]);
	free(object);
#endif

}

/**
 * Initializates the touchscreen subsystem
 *
 * @return A tsdev struct which can be used to read screen touchs
 */

struct tsdev *ttos_init_touchscreen() {

	struct tsdev *device;

#ifdef TTOS_USE_FAKE_FB
	device=(struct tsdev *)malloc(sizeof(struct tsdev));
	if (!device) {
		return (NULL);
	}
	device->fd=pipefd[0]; /* this must be the READ pipe */
	ts_rotation=rotation;
#endif

#ifdef TTOS_USE_SDL

	device=(struct tsdev *)malloc(sizeof(struct tsdev));
	if (!device) {
		return (NULL);
	}
	pipe(pipefd);

	device->fd=pipefd[0]; /* this must be the READ pipe */
	pthread_create(&ts_thread,NULL,ttos_emulate_ts_SDL,NULL);
	ts_rotation=rotation;
#endif

#ifdef TTOS_USE_TSLIB
	char *tsdevice=NULL;

	if( (tsdevice = getenv("TSLIB_TSDEVICE")) != NULL ) {
		device = ts_open(tsdevice,0);
	} else {
		if (!(device = ts_open("/dev/input/event0", 0)))
			device = ts_open("/dev/touchscreen/ucb1x00", 0);
	}

	if (!device) {
		return NULL;
	}

	if (ts_config(device)) {
		return NULL;
	}
	ts_rotation=0;
#endif

	ts_status=TTOS_TOUCH_NONE;
	ts_x=0;
	ts_y=0;

	return (device);
}

#ifdef TTOS_USE_FAKE_FB

int ts_fd(struct tsdev *ts) {
	return ts->fd;
}

#endif

#ifdef TTOS_USE_SDL

int ts_fd(struct tsdev *ts) {
	return ts->fd;
}

#endif

/**
 * Reads a touch in the screen, returning a sample with the coordinates
 *
 * @param ts The struct returned by ttos_init_touchscreen
 * @param sample A pointer to a struct where to store the touch
 *
 * @return 0 on success; -1 on failure
 */

int32_t ttos_read_ts(struct tsdev *ts, struct ttos_tssample *sample) {

	int x,y,type;

#ifdef TTOS_USE_FAKE_FB
	sample->type=TTOS_TOUCH_NONE;
	gettimeofday(&sample->tv,NULL);
	return 0;
#endif

	if ((ts==NULL)||(sample==NULL))
		return -1;

#ifdef TTOS_USE_SDL

	unsigned char data[5];
	read(ts->fd,data,5);
	x=(int)(data[0])+256*((int)data[1]);
	y=(int)(data[2])+256*((int)data[3]);
	type=data[4];
	gettimeofday(&sample->tv,NULL);
#endif

#ifdef TTOS_USE_TSLIB
	struct ts_sample sample2;

	ts_read(ts,&sample2,1);
	x=sample2.x;
	y=sample2.y;
	type=sample2.pressure > 127 ? 1 : 0;
	sample->tv.tv_sec=sample2.tv.tv_sec;
	sample->tv.tv_usec=sample2.tv.tv_usec;
#endif

	switch(ts_rotation) {
	case 0:
		sample->x=x;
		sample->y=y;
	break;
	case 1:
		sample->x=y;
		sample->y=s_width-x;
	break;
	case 2:
		sample->x=s_width-x;
		sample->y=s_height-y;
	break;
	case 3:
		sample->x=s_height-y;
		sample->y=x;
	break;
	}

	sample->ox=ts_x;
	sample->oy=ts_y;

	switch(ts_status) {
	case TTOS_TOUCH_NONE:
		if (type) {
			sample->type=TTOS_TOUCH_START_CLICK;
			ts_x=sample->x;
			ts_y=sample->y;
			sample->ox=sample->x;
			sample->oy=sample->y;
			ts_status=TTOS_TOUCH_START_CLICK;
		}
	break;
	case TTOS_TOUCH_START_CLICK:
		if (!type) {
			sample->type=TTOS_TOUCH_CLICK;
			ts_status=TTOS_TOUCH_NONE;
		} else {
			if (((sample->ox-sample->x)>MAX_DANDD)||((sample->x-sample->ox)>MAX_DANDD)||((sample->oy-sample->y)>MAX_DANDD)||((sample->y-sample->oy)>MAX_DANDD)) {
				sample->type=TTOS_TOUCH_DRAG;
				ts_status=TTOS_TOUCH_DRAG;
			} else {
				sample->type=TTOS_TOUCH_NONE;
			}
		}
	break;
	case TTOS_TOUCH_DRAG:
		if (type) {
			sample->type=TTOS_TOUCH_DRAG;
		} else {
			sample->type=TTOS_TOUCH_END_DRAG;
			ts_status=TTOS_TOUCH_NONE;
		}
	break;
	default:
	break;
	}

	return 0;
}

