/*
 * Copyright 2011 Raster Software Vigo
 *
 * This file is part of ttos
 *
 * ttos is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * ttos is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

/*
 * helper_functions.c
 *
 *  Created on: 25/01/2011
 *      Author: raster
 */

/** @file */

#include <stdlib.h>
#include <stdio.h>
#include <sys/types.h>
#include <dirent.h>
#include <string.h>
#include <unistd.h>
#include "ttos.h"
#include "helper_functions.h"
#include <sys/stat.h>
#include <locale.h>
#include <libintl.h>

#define _(string) gettext (string)

extern char **environ;

/**
 * Callback for receiving all the messages sent by the window server through port 0
 *
 * @param port Receives the port number (0 in this case)
 * @param answer_port The desired port to send an answer (none in this case)
 * @param connection The minibus connection from which the message was received
 * @param user_data The local data; in this case, the global structure of this session
 *
 */

uint32_t ttos_main_callback(uint32_t port, uint32_t answer_port, struct minibus_connection *connection,void *user_data) {

	struct minibus_data_element data;
	struct minibus_data_element *params;
	struct ttos_object *master;
	int32_t error;
	uint32_t command;

	master=(struct ttos_object *)user_data;

	if (0!=minibus_read_data(connection,&data)) {
		return -1;
	}

	if (data.type!=MINIBUS_TYPE_INT32) {
		return -1;
	}

	command=data.value32;
	switch (command) {
	case TTOS_SR_INIT:
		params=minibus_read_params(connection,"llllbbblbblbblbbl",&error);
		if (!params) {
			return -1;
		}
		master->init_flags=params[0].value32;
		master->uid=params[1].value32;
		master->ipc_uid=params[2].value32;
		master->pixel_format=params[3].value32;
		master->bytes_per_pixel=params[4].value8;
		master->R_format.rshift=params[5].value8;
		master->R_format.lshift=params[6].value8;
		master->R_format.mask  =params[7].value32;
		master->G_format.rshift=params[8].value8;
		master->G_format.lshift=params[9].value8;
		master->G_format.mask  =params[10].value32;
		master->B_format.rshift=params[11].value8;
		master->B_format.lshift=params[12].value8;
		master->B_format.mask  =params[13].value32;
		master->A_format.rshift=params[14].value8;
		master->A_format.lshift=params[15].value8;
		master->A_format.mask  =params[16].value32;
		free(params);
	break;
	case TTOS_SR_NEW_RESOLUTION:
		params=minibus_read_params(connection,"ssss",&error);
		if (!params) {
			return -1;
		}
		master->ox=params[0].value16;
		master->oy=params[1].value16;
		master->width=params[2].value16;
		master->height=params[3].value16;
		free(params);
	break;
	case TTOS_SR_MOVED_TO_FG:
		master->fg=true;
	break;
	case TTOS_SR_MOVED_TO_BG:
		master->fg=false;
	break;
	case TTOS_SR_NEW_CLIENT:
		params=minibus_read_params(connection,"lc",&error);
		if (!params) {
			return -1;
		}
		if (master->new_client) {
			master->new_client(master,params[0].value32,params[1].string,master->new_client_data);
		}
		free(params);
	break;
	case TTOS_SR_DESTROYED_CLIENT:
		params=minibus_read_params(connection,"lbb",&error);
		if (!params) {
			return -1;
		}
		if (master->destroyed_client) {
			master->destroyed_client(master,params[0].value32,params[2].value8,params[1].value8,master->destroyed_client_data);
		}
		free(params);
	break;
	case TTOS_SR_EXIT:
		if (master->exit_cb) {
			master->exit_cb(master,master->exit_cb_data);
		}
		master->conn->master_queue->forced_exit=true;
		minibus_exit_loop(master->conn->master_queue);
	break;
	}

	if ((master->cb_port0)&&(command!=TTOS_SR_NEW_CLIENT)&&(command!=TTOS_SR_DESTROYED_CLIENT)&&(port)) {
		master->cb_port0(master,command,master->cb_port0_data);
	}

	return 0;
}


/**
 * Waits for an error in a port, and returns it. While waiting, all callbacks will be called
 *
 * @param object The server object
 * @param port The port
 * @param error A pointer to an integer where the received error will be stored
 * @return A TTOS_SYSTEM_ERRORS error
 */

uint32_t ttos_wait_for_error_2(struct ttos_object *object, uint32_t port,uint32_t *error) {

	struct minibus_data_element *params;
	int32_t error2;

	if (!object) {
		return -1;
	}

	minibus_wait_for_port(object->conn,port,true);

	// Wait for three longs: first, the "answer port", second, the TTOS_SR_ERROR indicative, and third, the error itself
	params=minibus_read_params(object->conn,"lll",&error2);

	if ((!params)||(params[1].value32!=TTOS_SR_ERROR)) {
		free(params);
		return 1;
	}

	*error=params[2].value32;
	free(params);
	return 0;
}


/**
 * Waits for a TTOS_SR_NEW_WIDGET or TTOS_SR_ERROR message in a port, and returns its value; if the message isn't any of them, it
 * will return the message itself. Since widget UIDs, errors, and message types uses different number sets, is possible to
 * easily differentiate them.
 *
 * @param object The server object
 * @param port The port
  *
 * @return If <0x80000000, is an UID; if >=0x80000000 and <0xC0000000, is a message type; else, is a A TTOS_SYSTEM_ERRORS error
 */

uint32_t ttos_wait_for_error(struct ttos_object *object, uint32_t port) {

	struct minibus_data_element *params;
	int32_t error2;
	uint32_t error,command;

	if (!object) {
		return TTOS_ERROR_CANT_READ_DATA;
	}

	minibus_wait_for_port(object->conn,port,false);

	// Wait for three longs: first, the "answer port", second, the TTOS_SR_ERROR indicative, and third, the error itself
	params=minibus_read_params(object->conn,"ll",&error2);

	if (!params) {
		return TTOS_ERROR_INCORRECT_ANSWER_FORMAT;
	}
	command=params[1].value32;
	free(params);

	if (command<0x80000000) {
		return TTOS_ERROR_INCORRECT_ANSWER_FORMAT;
	}

	if ((command==TTOS_SR_ERROR)||(command==TTOS_SR_NEW_WIDGET)) {
		// here we read the error or UID itself
		params=minibus_read_params(object->conn,"l",&error2);
		if (!params){
			return TTOS_ERROR_INCORRECT_ANSWER_FORMAT;
		}
		error=params[0].value32;
		free(params);
		ttos_print_error(error,false);
		return(error);
	}
	return(command);
}


struct _ttos_select_file_object *ttos_new_select_file_object(Cstring name, bool folder, uint32_t widget_uid) {

	struct _ttos_select_file_object *object;

	object=malloc(sizeof(struct _ttos_select_file_object));
	if (!object) {
		return (NULL);
	}
	object->folder=folder;
	object->widget_uid=widget_uid;
	object->name=strdup(name);
	return (object);
}

void ttos_destroy_select_file_object(void *object) {

	struct _ttos_select_file_object *selector;

	if (!object) {
		return;
	}

	selector=object;
	if (selector->name) {
		free(selector->name);
	}
	free(selector);
}

bool ttos_select_file_comparer(void *v1, void *v2) {

	struct _ttos_select_file_object *file1,*file2;

	if (!v1) {
		return true;
	}

	if (!v2) {
		return false;
	}
	file1=v1;
	file2=v2;
	if (file1->folder!=file2->folder) {
		return (file1->folder);
	}
	if (strcasecmp(file1->name,file2->name)>0) {
		return false;
	} else {
		return true;
	}
}

/**
 * Shows a file selector in a new tab. When it returns, the tab remains, so the caller can decide, based on the answer, if
 * it wants to change to a new tab, exit... and how to change. The tab created here is marked to be deleted automatically.
 *
 * @param master The connection to the compositor
 * @param base_path The first path to show
 * @param flags TTOS_FILE_SHOW_HIDDEN: shows the hidden files; TTOS_FILE_SELECT_FOLDER show only folders, not files; TTOS_FILE_ALLOW_GO_UP allow to go to folders located upstream the initial one
 * @param filetypes An array of enum ttos_file_type elements with the file types to be shown, ended in TTOS_FTYPE_END_LIST, or NULL to show all file types
 *
 * @return A Cstring with the selected file/folder, or NULL if the user cancelled (must be freed by the caller)
 */


Cstring ttos_select_file(struct ttos_object *master, Cstring base_path, uint32_t flags, enum ttos_file_type *filetypes) {

	uint32_t current_tab;
	uint32_t old_tab;
	int32_t error;
	uint32_t uid;
	uint32_t width;
	Cstring current_path,upper_path,tail_path,icon,tmp_path,retval;
	DIR *folder;
	struct ttos_color color;
	struct minibus_list files;
	struct dirent *file;
	struct _ttos_select_file_object *newfile;
	struct minibus_list_element *element;
	struct minibus_data_element *data;
	bool is_folder;
	bool scroll_left;
	uint16_t y;
	uint32_t back,ok,cancel,t_uid;
	bool do_select;
	enum ttos_file_type ftype,*fpointer;
	uint32_t rv;

	current_path=strdup(base_path);
	old_tab=0;
	color.r=color.g=color.b=color.a=0xFF;

	t_uid=ttos_get_visible_tab(master,TTOS_SYNC_ANSWER);

	minibus_list_init(&files,ttos_destroy_select_file_object,ttos_select_file_comparer);

	retval=NULL;
	do_select=true;
	do {
		back=ok=cancel=0;
		folder=opendir(current_path);
		if(!folder) {
			break;
		}
		minibus_list_free(&files);
		while(NULL!=(file=readdir(folder))) {
			if (*file->d_name=='.') {
				if (!(flags&TTOS_FILE_SHOW_HIDDEN)) {
					continue;
				}
				if ((!strcmp(file->d_name,"."))||(!strcmp(file->d_name,".."))) {
					continue;
				}
			}
			switch(file->d_type) {
			case DT_DIR:
				is_folder=true;
			break;
			case DT_REG:
				is_folder=false;
			break;
			default:
				continue;
			break;
			}
			newfile=ttos_new_select_file_object(file->d_name,is_folder,0);
			if (!newfile) {
				break;
			}
			minibus_list_add_sorted(&files,newfile);
		}
		closedir(folder);

		if(file) { // if it's not NULL, there was an error
			break;
		}

		current_tab=ttos_new_tab(master,TTOS_SYNC_ANSWER);
		if (current_tab>TTOS_ERROR_NO_ERROR) { // an error ocurred
			break;
		}
		color.r=color.g=color.b=color.a=0xFF;
		rv=ttos_set_tab_bg(master,TTOS_SYNC_ANSWER,current_tab,&color);
		color.b=0xBF;
		color.g=color.r=color.a=0xFF;

		y=49;

		ttos_set_tab_margins(master,TTOS_SYNC_ANSWER,current_tab,0,5,0,0);

		for(element=files.list;element;element=element->next) {
			newfile=element->data;
			if (newfile->folder) {
				icon="folder.tga";
			} else if (!(flags&TTOS_FILE_SELECT_FOLDER)) {
				ftype=ttos_check_extension(newfile->name);
				if (filetypes) {
					for(fpointer=filetypes;*fpointer!=TTOS_FTYPE_END_LIST;fpointer++) {
						if (ftype==*fpointer) {
							break;
						}
					}
					if (*fpointer==TTOS_FTYPE_END_LIST) {
						continue;
					}
				}
				switch (ftype) {
				case TTOS_FTYPE_MP3:
				case TTOS_FTYPE_OGG:
				case TTOS_FTYPE_VORBIS:
					icon="sound.tga";
				break;
				default:
					icon="anyfile.tga";
				break;
				}
			} else {
				continue;
			}
			newfile->widget_uid=ttos_new_button(master,TTOS_SYNC_ANSWER,current_tab,5,y,master->width-10,34,TTOS_NO_STOCK_BUTTON,icon,newfile->name,NULL,NULL,TTOS_WIDGET_FONT_HALIGN_LEFT|TTOS_WIDGET_FONT_VALIGN_MIDDLE,TTOS_RESERVED_PORTS+1);
			y+=36;
		}


		ttos_new_color_area(master,TTOS_SYNC_ANSWER,current_tab,0,0,master->width,44,&color,TTOS_WIDGET_FIXED,TTOS_NO_ANSWER);
		upper_path=ttos_get_upper_path(current_path,&tail_path);
		if(!tail_path) {
			tail_path=current_path;
		}
		width = (flags&TTOS_FILE_SELECT_FOLDER) ? 126 : 88;
		ttos_new_label(master,TTOS_SYNC_ANSWER,current_tab,44,5,master->width-width,34,tail_path,NULL,TTOS_WIDGET_FONT_HALIGN_CENTER|TTOS_WIDGET_FONT_VALIGN_MIDDLE|TTOS_WIDGET_FIXED,TTOS_NO_ANSWER);
		if ((strcmp(current_path,base_path))||(flags&TTOS_FILE_ALLOW_GO_UP)) {
			back=ttos_new_button(master,TTOS_SYNC_ANSWER,current_tab,5,5,34,34,TTOS_LEFT_BUTTON,NULL,NULL,NULL,NULL,TTOS_WIDGET_FONT_HALIGN_CENTER|TTOS_WIDGET_FONT_VALIGN_MIDDLE|TTOS_WIDGET_FIXED|TTOS_WIDGET_BUTTON_NO_TEXT,TTOS_RESERVED_PORTS+1);
		}
		cancel=ttos_new_button(master,TTOS_SYNC_ANSWER,current_tab,master->width-39,5,34,34,TTOS_CANCEL_BUTTON,NULL,NULL,NULL,NULL,TTOS_WIDGET_FONT_HALIGN_CENTER|TTOS_WIDGET_FONT_VALIGN_MIDDLE|TTOS_WIDGET_FIXED|TTOS_WIDGET_BUTTON_NO_TEXT,TTOS_RESERVED_PORTS+1);
		if (flags&TTOS_FILE_SELECT_FOLDER) {
			ok=ttos_new_button(master,TTOS_SYNC_ANSWER,current_tab,master->width-78,5,34,34,TTOS_OK_BUTTON,NULL,NULL,NULL,NULL,TTOS_WIDGET_FONT_HALIGN_CENTER|TTOS_WIDGET_FONT_VALIGN_MIDDLE|TTOS_WIDGET_FIXED|TTOS_WIDGET_BUTTON_NO_TEXT,TTOS_RESERVED_PORTS+1);
		}

		if (!old_tab) {
			rv=ttos_set_visible_tab(master,TTOS_SYNC_ANSWER,current_tab,TTOS_TAB_SWAP_MODE_SCROLL_UP, TTOS_RESERVED_PORTS);
			if ((TTOS_ERROR_NO_ERROR==rv)||(TTOS_ERROR_DESTROY_DELAYED==rv)) {
				minibus_wait_for_port(master->conn,TTOS_RESERVED_PORTS,false);
			} else {
				ttos_print_error(rv,true);
			}
			ttos_repaint_screen(master,TTOS_SYNC_ANSWER);
		} else {
			rv=ttos_set_visible_tab(master,TTOS_SYNC_ANSWER,current_tab,scroll_left ? TTOS_TAB_SWAP_MODE_SCROLL_LEFT : TTOS_TAB_SWAP_MODE_SCROLL_RIGHT, TTOS_RESERVED_PORTS);
			if ((TTOS_ERROR_NO_ERROR==rv)||(TTOS_ERROR_DESTROY_DELAYED==rv)) {
				minibus_wait_for_port(master->conn,TTOS_RESERVED_PORTS,false);
			} else {
				ttos_print_error(rv,true);
			}
			ttos_repaint_screen(master,TTOS_SYNC_ANSWER);
			ttos_destroy_tab(master,TTOS_SYNC_ANSWER,old_tab,0);
		}
		old_tab=current_tab;
		do {
			minibus_wait_for_port(master->conn,TTOS_RESERVED_PORTS+1,false);
			if (master->conn->master_queue->forced_exit) { // if a forced exit has been received, just cancel
				break;
			}
			data=minibus_read_params(master->conn,"llblllll",&error);
		} while (!data);

		if (master->conn->master_queue->forced_exit) {
			uid=cancel;
		} else {
			uid=data[3].value32;
			free(data);
		}

		if (uid==back) {
			free(current_path);
			current_path=upper_path;
			scroll_left=false;
			continue;
		}
		free(upper_path);
		if (uid==cancel) {
			free(current_path);
			break;
		}
		if ((flags&TTOS_FILE_SELECT_FOLDER) && (uid==ok)) {
			retval=current_path;
			break;
		}
		for(element=files.list;element;element=element->next) {
			newfile=element->data;
			if(newfile->widget_uid==uid) {
				if (newfile->folder) {
					tmp_path=minibus_path_join(current_path,newfile->name);
					free(current_path);
					current_path=tmp_path;
					scroll_left=true;
					break;
				}
				retval=minibus_path_join(current_path,newfile->name);
				free(current_path);
				do_select=false;
				break;
			}
		}
	} while(do_select);

	minibus_list_free(&files);

	if (old_tab) {
		ttos_destroy_tab(master,TTOS_SYNC_ANSWER,old_tab,TTOS_TAB_DELAY_DESTROY);
	}

	return (retval);

}


/**
 * Shows a dialog text with one or more buttons for the user. When it returns, the tab remains, so the caller can decide, based
 * on the answer, if it wants to change to a new tab, exit... and how to change. The tab created here is marked to be deleted
 * automatically.
 *
 * @param master The connection to the compositor
 * @param text The text to be shown in the dialog text
 * @param text_flags Text flags (TTOS_WIDGET_FONT_MULTILINE, TTOS_WIDGET_FONT_WORD_CUT, TTOS_WIDGET_FONT_LATIN1,
 * TTOS_WIDGET_FONT_HALIGN_LEFT, TTOS_WIDGET_FONT_HALIGN_CENTER, TTOS_WIDGET_FONT_HALIGN_RIGHT, TTOS_WIDGET_FONT_VALIGN_TOP,
 * TTOS_WIDGET_FONT_VALIGN_MIDDLE, TTOS_WIDGET_FONT_VALIGN_BOTTOM and TTOS_WIDGET_FONT_ALIGN_CENTER)
 * @param buttons The list of buttons to show, from left to right
 * @param count Number of buttons in the list
 *
 * @return The code of the pressed button, or -1 on error
 */


int32_t ttos_show_dialog(struct ttos_object *master, Cstring text, uint32_t text_flags, struct ttos_dialog_button *buttons, uint16_t count) {

	int16_t inc,w;
	uint16_t loop;
	uint32_t *uids,current_tab,r_uid,t_uid;
	int32_t error;
	struct ttos_color color;
	struct minibus_data_element *data;

	if ((!master)||(!text)||(!buttons)||(!count)) {
		return -1;
	}

	current_tab=ttos_new_tab(master,TTOS_SYNC_ANSWER);
	if (current_tab>TTOS_ERROR_NO_ERROR) { // an error ocurred
		return -1;
	}

	t_uid=ttos_get_visible_tab(master,TTOS_SYNC_ANSWER);

	uids=malloc(sizeof(uint32_t)*count);
	if (!uids) {
		return -1;
	}

	ttos_new_label(master,TTOS_SYNC_ANSWER,current_tab,5,5,master->width-10,master->height-45,text,NULL,text_flags,TTOS_NO_ANSWER);
	color.r=color.g=color.b=color.a=0xFF;
	ttos_set_tab_bg(master,TTOS_SYNC_ANSWER,current_tab,&color);
	w=(master->width-5)/count-5;
	for(loop=0;loop<count;loop++) {
		inc=((master->width-5)*loop)/count;
		uids[loop]=ttos_new_button(master,TTOS_SYNC_ANSWER,current_tab,5+inc,master->height-43,w,38,buttons[loop].button,buttons[loop].icon,buttons[loop].text,NULL,NULL,TTOS_WIDGET_FONT_ALIGN_CENTER,TTOS_RESERVED_PORTS+2);
	}
	ttos_set_visible_tab(master,TTOS_SYNC_ANSWER,current_tab,TTOS_TAB_SWAP_MODE_REPLACE, TTOS_NO_ANSWER);
	ttos_destroy_tab(master,TTOS_SYNC_ANSWER,current_tab,TTOS_TAB_DELAY_DESTROY);
	ttos_repaint_screen(master,TTOS_SYNC_ANSWER);
	do {
		minibus_wait_for_port(master->conn,TTOS_RESERVED_PORTS+2,false);
		if (master->conn->master_queue->forced_exit) {
			data=NULL;
			break;
		}
		data=minibus_read_params(master->conn,"llblllll",&error);
	} while (!data);
	if (data) {
		r_uid=data[3].value32;
		free(data);
	} else {
		r_uid=0;
	}

	for(loop=0;loop<count;loop++) {
		if (r_uid==uids[loop]) {
			free(uids);
			return ((int32_t)buttons[loop].return_value);
		}
	}
	free(uids);
	return -1;
}


/**
 * Launchs a program in parallel, as a child. It searchs the binary in $TTOS_BASE_PATH/programs, $TTOS_BASE_PATH, ./programs,
 * /bin, /usr/bin and /usr/local/bin
 *
 * @param command An array of argument strings passed to the new program, ended in NULL
 * @param pipes Two pipe UIDs, which will be connected, the first to STDIN, and the second to STDOUT and STDERR. If NULL, the STD pipes won't be redirected
 *
 * @return The child's PID on success, or -1 on error (a success doesn't mean full success; it still can fail the launch
 * of the program itself).
 */


int32_t ttos_launch_child(Cstring command[], int *pipes) {

	Cstring *final_command,program,final_program;
	char *current_path;
	uint32_t counter,total;
	int32_t child_pid;

	if ((!command)||(!(*command[0]))) {
		return -1;
	}

	child_pid=fork();
	if(child_pid) {
		return (child_pid);
	}

	setsid();
	if (pipes) {
		dup2(pipes[0],0);
		dup2(pipes[1],1);
		dup2(pipes[1],2);
	}

	for(total=0;command[total];total++);
	total++;
	final_command=malloc(sizeof(Cstring)*total);
	if (!final_command)
		exit(127);

	for(counter=0;counter<total;counter++) {
		final_command[counter]=command[counter];
	}

	program=final_command[0];

	// try with "TTOS_BASE_PATH/programs"
	current_path=getenv("TTOS_BASE_PATH");
	if ((current_path)&&(current_path[0]!=0)) {
		final_program=minibus_path_join(current_path,"programs");
		final_command[0]=minibus_path_join(final_program,program);
		execve(final_command[0],final_command,environ);
		free(final_program);
		free(final_command[0]);
		// and with TTOS_BASE_PATH, just in case
		final_command[0]=minibus_path_join(current_path,program);
		execve(final_command[0],final_command,environ);
		free(final_command[0]);
	}

	// If we reached this point, that file didn't exists,
	// so let's try with "current_path/programs"

	current_path=getcwd(NULL,0);
	final_program=minibus_path_join(current_path,"programs");
	final_command[0]=minibus_path_join(final_program,program);
	execve(final_command[0],final_command,environ);
	free(current_path);
	free(final_program);
	free(final_command[0]);

	// If we reached this point, that file didn't exists,
	// so let's try /usr/bin

	final_command[0]=minibus_path_join("/usr/bin",program);
	execve(final_command[0],final_command,environ);
	free(final_command[0]);

	// If we reached this point, that file didn't exists,
	// so let's try /usr/local/bin

	final_command[0]=minibus_path_join("/usr/local/bin",program);
	execve(final_command[0],final_command,environ);
	free(final_command[0]);

	// If we reached this point, that file didn't exists,
	// so let's try /bin

	final_command[0]=minibus_path_join("/bin",program);
	execve(final_command[0],final_command,environ);
	free(final_command[0]);

	// The program don't exist. Return an error value

	exit(127);
}

/**
 * Paints a grid of buttons in the specified tab, and, optionally, returns an array with their UIDs. It can be horizontal (starting
 * at MARGIN from left, and extending as long as needed to the right, and putting as many as possible in vertical) or vertical
 * (starting at MARGIN from top, and extending as long as needed to bottom, and putting as many as possible in horizontal).
 *
 * @param master The connection to the compositor
 * @param tab The tab where to paint the buttons
 * @param margin Margin at left or up (to allow to put labels, fixed buttons, and so on)
 * @param vertical If true, the grid will extend vertically; if false, it will extend horizontally
 * @param elements Number of elements
 * @param buttons An array of TTOS_KEYBOARD elements, describing the buttons to paint. Width and Height means the desired button size.
 * @param return_uids If true, the function will return an array with the UID of each button. This array must be freed by the caller.
 *
 * @return The button uids if desired, or 1 if all went fine; or NULL if there was an error.
 */

uint32_t *ttos_paint_grid_buttons(struct ttos_object *master, uint32_t tab, int16_t margin, bool vertical, uint16_t elements,
		struct ttos_keyboard *buttons, bool return_uids) {

	uint32_t margin2;
	int16_t x;
	int16_t y;
	int16_t he,ve;
	uint32_t counter,sw,sh,loop,kw,kh;
	uint32_t *uids,tmp,*p2,flags;
	struct _ttos_key_buttons *pointer;

	if (return_uids) {
		uids=malloc(sizeof(uint32_t)*elements);
		if (!uids) {
			return NULL;
		}
	} else {
		uids=NULL;
	}
	p2=uids;

	kw=buttons->width;
	kh=buttons->height;
	pointer=buttons->buttons;

	if (vertical) {
		he=master->width/kw;
		margin2=(master->width%kw)/(1+he);
		ttos_set_tab_margins(master,TTOS_SYNC_ANSWER,tab,0,margin2,0,0);
		y=margin+margin2;
		sw=master->width-margin2;

		for(counter=0;counter<elements;) {
			for(loop=0;loop<he;loop++) {
				x=(int16_t)(margin2+((loop*sw)/he));
				if (pointer->text) {
					flags=TTOS_WIDGET_FONT_HALIGN_CENTER|TTOS_WIDGET_FONT_VALIGN_MIDDLE;
				} else {
					flags=TTOS_WIDGET_FONT_HALIGN_CENTER|TTOS_WIDGET_FONT_VALIGN_MIDDLE|TTOS_WIDGET_BUTTON_NO_TEXT;
				}
				if ((pointer->string!=NULL)||(pointer->icon!=NULL)||(pointer->stock_button!=TTOS_NO_STOCK_BUTTON)) {
					tmp=ttos_new_button(master,TTOS_SYNC_ANSWER,tab,x,y,kw,kh,pointer->stock_button,pointer->icon,_(pointer->string),&buttons->colors[pointer->color],NULL,flags,pointer->event);
				} else {
					tmp=0;
				}
				if (return_uids) {
					*(p2++)=tmp;
				}
				counter++;
				if (counter>=elements) {
					break;
				}
				pointer++;
			}
			y+=kh+margin2;
		}
	} else {
		printf("Horizontal\n");
		ve=master->height/kh;
		margin2=(master->height%kh)/(1+ve);
		x=margin+margin2;
		sh=master->height-margin2;
		ttos_set_tab_margins(master,TTOS_SYNC_ANSWER,tab,0,0,0,margin2);

		for(counter=0;counter<elements;) {
			for(loop=0;loop<ve;loop++) {
				y=(int16_t)(margin2+((loop*sh)/ve));
				if (pointer->text) {
					flags=TTOS_WIDGET_FONT_HALIGN_CENTER|TTOS_WIDGET_FONT_VALIGN_MIDDLE;
				} else {
					flags=TTOS_WIDGET_FONT_HALIGN_CENTER|TTOS_WIDGET_FONT_VALIGN_MIDDLE|TTOS_WIDGET_BUTTON_NO_TEXT;
				}
				if ((pointer->string!=NULL)||(pointer->icon!=NULL)||(pointer->stock_button!=TTOS_NO_STOCK_BUTTON)) {
					tmp=ttos_new_button(master,TTOS_SYNC_ANSWER,tab,x,y,kw,kh,pointer->stock_button,pointer->icon,_(pointer->string),&buttons->colors[pointer->color],NULL,flags,pointer->event);
				} else {
					tmp=0;
				}
				if (return_uids) {
					*(p2++)=tmp;
				}
				counter++;
				if (counter>=elements) {
					break;
				}
				pointer++;
			}
			x+=kw+margin2;
		}
	}

	if (uids) {
		return (uids);
	} else {
		return ((uint32_t *)1);
	}
}

/**
 * Paints a grid of buttons in the specified tab, and, optionally, returns an array with their UIDs. The size of each button
 * will be automatically calculated to ensure that the grid fills the desired tab zone, forming a keyboard.
 *
 * @param master The connection to the compositor
 * @param tab The tab where to paint the buttons
 * @param ox The X coordinate of the rectangle to fill with the keyboard
 * @param oy The Y coordinate of the rectangle to fill with the keyboard
 * @param width The width of the rectangle to fill with the keyboard
 * @param height The height of the rectangle to fill with the keyboard
 * @param h_margin Horizontal margin between buttons
 * @param v_margin Vertical margin between buttons
 * @param keyboard An array of buttons. Width and Height are, respectively, the number of columns and rows.
 * @param return_uids If true, it will return an array with all the buttons'UIDs. This array must be freed by the caller.
 *
 * @return The button uids if desired, or 1 if all went fine; or NULL if there was an error.
 */

uint32_t *ttos_paint_keyboard(struct ttos_object *master, uint32_t tab, int16_t ox, int16_t oy, int16_t width, int16_t height,
		int16_t h_margin, int16_t v_margin, struct ttos_keyboard *keyboard, bool return_uids) {

	int32_t w2,h2;
	int32_t x,y;
	uint32_t loop1,loop2,*uids,*p2,tmp;
	struct _ttos_key_buttons *pointer;
	uint32_t key_w;
	uint32_t key_h,flags;

	key_w=keyboard->width;
	key_h=keyboard->height;

	if (return_uids) {
		uids=malloc(4*key_w*key_h);
		if (!uids) {
			return NULL;
		}
		p2=uids;
	} else {
		uids=NULL;
	}

	pointer=keyboard->buttons;
	width+=h_margin;
	height+=v_margin;

	for(loop2=0;loop2<key_h;loop2++) {
		y=(height*loop2)/key_h;
		for(loop1=0;loop1<key_w;loop1++) {
			x=(width*loop1)/key_w;
			if (pointer->size&0x01) {
				w2=((width*(loop1+2))/key_w)-x-h_margin;
			} else {
				w2=((width*(loop1+1))/key_w)-x-h_margin;
			}
			if (pointer->size&0x02) {
				h2=((height*(loop2+2))/key_h)-y-v_margin;
			} else {
				h2=((height*(loop2+1))/key_h)-y-v_margin;
			}
			if (pointer->text) {
				flags=TTOS_WIDGET_FONT_HALIGN_CENTER|TTOS_WIDGET_FONT_VALIGN_MIDDLE;
			} else {
				flags=TTOS_WIDGET_FONT_HALIGN_CENTER|TTOS_WIDGET_FONT_VALIGN_MIDDLE|TTOS_WIDGET_BUTTON_NO_TEXT;
			}
			if ((pointer->string!=NULL)||(pointer->icon!=NULL)||(pointer->stock_button!=TTOS_NO_STOCK_BUTTON)) {
				tmp=ttos_new_button(master,TTOS_SYNC_ANSWER,tab,x+ox,y+oy,w2,h2,pointer->stock_button,pointer->icon,pointer->string,&keyboard->colors[pointer->color],NULL,flags,pointer->event);
			} else {
				tmp=0;
			}
			if (return_uids) {
				*(p2++)=tmp;
			}
			pointer++;
		}
	}

	return uids;
}


/**
 * Prints the type and content of a MINIBUS_DATA_ELEMENT. Useful for debugging
 *
 * @param data The data to print
 */

void ttos_print_type(struct minibus_data_element *data) {

	switch (data->type) {
		case MINIBUS_TYPE_INT8:
			printf("uint8_t %X ",data->value8);
		break;
		case MINIBUS_TYPE_INT16:
			printf("uint16_t %X ",data->value16);
		break;
		case MINIBUS_TYPE_INT32:
			printf("uint32_t %X ",data->value32);
		break;
		case MINIBUS_TYPE_STRING:
			printf("string (%d) '%s' ",data->stringsize,data->string);
		break;
		case MINIBUS_TYPE_BLOB:
			printf("BLOB (%d) '%s' ",data->blobsize,data->blob);
		break;
		default:
		break;
	}
}

/**
 * Prints the event specified. Useful for debugging
 *
 * @param event The event to print
 */

void ttos_print_event(uint32_t event) {

	switch (event) {
	case TTOS_SR_INIT:
		printf("TTOS_SR_INIT");
		break;
	case TTOS_SR_ERROR:
		printf("TTOS_SR_ERROR");
		break;
	case TTOS_SR_NEW_RESOLUTION:
		printf("TTOS_SR_NEW_RESOLUTION");
		break;
	case TTOS_SR_NEW_CLIENT:
		printf("TTOS_SR_NEW_CLIENT");
		break;
	case TTOS_SR_DESTROYED_CLIENT:
		printf("TTOS_SR_DESTROYED_CLIENT");
		break;
	case TTOS_SR_MOVED_TO_BG:
		printf("TTOS_SR_MOVED_TO_BG");
		break;
	case TTOS_SR_MOVED_TO_FG:
		printf("TTOS_SR_MOVED_TO_FG");
		break;
	case TTOS_SR_TOUCH_EVENT:
		printf("TTOS_SR_TOUCH_EVENT");
		break;
	case TTOS_SR_EXIT:
		printf("TTOS_SR_EXIT");
		break;
	case TTOS_SR_TIMER:
		printf("TTOS_SR_TIMER");
		break;
	case TTOS_SR_SCROLL:
		printf("TTOS_SR_SCROLL");
		break;
	case TTOS_SR_TAB_DRAG:
		printf("TTOS_SR_TAB_DRAG");
		break;
	case TTOS_SR_NEW_WIDGET:
		printf("TTOS_SR_NEW_WIDGET");
		break;
	default:
		printf("Unknown event %d", event);
		break;
	}
}

/**
 * Prints the specified error. Useful for debugging
 *
 * @param error The error to print
 * @param print_no_error If true, also print the error NO_ERROR, and when the value is not an error
 */

void ttos_print_error(enum TTOS_SYSTEM_ERRORS error,bool print_no_error) {

	if (error<TTOS_ERROR_NO_ERROR) {
		if (print_no_error) {
			printf ("Not an error: %X\n",error);
		}
		return;
	}

	switch (error) {
	case TTOS_ERROR_NO_ERROR:
		if (print_no_error) {
			printf ("No error\n");
		}
	break;
	case TTOS_ERROR_DESTROY_DELAYED:
		printf("ERROR_DESTROY_DELAYED\n");
	break;
	case TTOS_ERROR_ASYNC_RESPONSE:
		printf ("ERROR_ASYNC_RESPONSE\n");
	break;
	case TTOS_ERROR_CANT_SEND_DATA:
		printf ("ERROR_CANT_SEND_DATA\n");
	break;
	case TTOS_ERROR_CANT_READ_DATA:
		printf ("ERROR_CANT_READ_DATA\n");
	break;
	case TTOS_ERROR_INCORRECT_FORMAT:
		printf ("ERROR_INCORRECT_FORMAT\n");
	break;
	case TTOS_ERROR_INCORRECT_ANSWER_FORMAT:
		printf ("ERROR_INCORRECT_ANSWER_FORMAT\n");
	break;
	case TTOS_ERROR_CANT_LOAD_PICTURE:
		printf ("ERROR_CANT_LOAD_PICTURE\n");
	break;
	case TTOS_ERROR_TOO_MANY_WIDGETS:
		printf ("ERROR_TOO_MANY_WIDGETS\n");
	break;
	case TTOS_ERROR_NO_MEMORY:
		printf ("ERROR_NO_MEMORY\n");
	break;
	case TTOS_ERROR_INVALID_WINDOW_SIZE:
		printf ("ERROR_INVALID_WINDOW_SIZE\n");
	break;
	case TTOS_ERROR_RESTRICTED_COMMAND:
		printf ("ERROR_RESTRICTED_COMMAND\n");
	break;
	case TTOS_ERROR_WINDOW_OUT_OF_RANGE:
		printf ("ERROR_WINDOW_OUT_OF_RANGE\n");
	break;
	case TTOS_ERROR_INVALID_UID:
		printf ("ERROR_INVALID_UID\n");
	break;
	case TTOS_ERROR_INVALID_TAB_UID:
		printf ("ERROR_INVALID_TAB_UID\n");
	break;
	case TTOS_ERROR_NO_VISIBLE_TAB:
		printf ("ERROR_NO_VISIBLE_TAB\n");
	break;
	case TTOS_ERROR_CANT_CREATE_TAB:
		printf ("ERROR_CANT_CREATE_TAB\n");
	break;
	case TTOS_ERROR_FILE_NOT_FOUND:
		printf ("ERROR_FILE_NOT_FOUND\n");
	break;
	case TTOS_ERROR_OLD_WINDOW_HAS_NO_FG:
		printf("ERROR_OLD_WINDOW_HAS_NO_FG\n");
	break;
	case TTOS_ERROR_INVALID_SWAP_MODE:
		printf("ERROR_INVALID_SWAP_MODE\n");
	break;
	case TTOS_ERROR_INCORRECT_BLOB_SIZE:
		printf("TTOS_ERROR_INCORRECT_BLOB_SIZE\n");
	break;
	case TTOS_ERROR_SHMEM_IN_USE:
		printf("ERROR_SHMEM_IN_USE\n");
	break;
	case TTOS_ERROR_INVALID_SHAREMEM_UID:
		printf("ERROR_INVALID_SHAREMEM_UID\n");
	break;
	case TTOS_ERROR_SHAREMEM_HAS_NO_ALPHA:
		printf("ERROR_SHAREMEM_HAS_NO_ALPHA\n");
	break;
	case TTOS_ERROR_CANT_SHARE_MEM:
		printf("ERROR_CANT_SHARE_MEM\n");
	break;
	case TTOS_ERROR_BUS_ALREADY_REGISTERED:
		printf("TTOS_ERROR_BUS_ALREADY_REGISTERED\n");
	break;
	case TTOS_ERROR_CANT_RESERVE_UID:
		printf("TTOS_ERROR_CANT_RESERVE_UID\n");
	break;
	case TTOS_ERROR_INVALID_BUS:
		printf("TTOS_ERROR_INVALID_BUS\n");
	break;
	case TTOS_ERROR_NO_OWNER:
		printf("TTOS_ERROR_NO_OWNER\n");
	break;
	case TTOS_ERROR_ALREADY_SUBSCRIBED:
		printf("TTOS_ERROR_ALREADY_SUBSCRIBED\n");
	break;
	case TTOS_ERROR_NOT_SUBSCRIBED:
		printf("TTOS_ERROR_NOT_SUBSCRIBED\n");
	break;
	case TTOS_ERROR_NOT_ALLOWED_TO_SEND:
		printf("TTOS_ERROR_NOT_ALLOWED_TO_SEND\n");
	break;
	case TTOS_ERROR_INVALID_IPC_PORT:
		printf("TTOS_ERROR_INVALID_IPC_PORT\n");
	break;
	case TTOS_ERROR_INVALID_IPC_ANSWER_PORT:
		printf("TTOS_ERROR_INVALID_IPC_ANSWER_PORT\n");
	break;
	case TTOS_ERROR_INVALID_IPC_MESSAGE_FORMAT:
		printf("TTOS_ERROR_INVALID_IPC_MESSAGE_FORMAT\n");
	break;
	default:
		printf("Unknown error %X\n",error);
	break;
	}
}

void ttos_destroy_registry_element(void *data) {

	struct ttos_registry_element *object=data;

	if (!data) {
		return;
	}

	if (object->value.type==MINIBUS_TYPE_STRING) {
		free(object->value.string);
	}
	if (object->value.type==MINIBUS_TYPE_BLOB) {
		free(object->value.blob);
	}
	free(object->key);
	free(object);
}

void ttos_jump_nl(FILE *fd) {

	char v;

	while(1==fread(&v,1,1,fd)) {
		if (v=='\n') {
			break;
		}
	}
}

int32_t ttos_read_blob(FILE *fd,uint8_t *buffer,uint32_t size) {

	uint32_t loop;
	uint8_t v,v1,v2;
	uint8_t *p;

	p=buffer;

	// jump over the first space
	if(1!=fread(&v,1,1,fd)) {
		return -1;
	}
	for(loop=0;loop<size;loop++) {
		if(1!=fread(&v,1,1,fd)) {
			return -1;
		}
		if (v<' ') { // avoid control characters
			return -1;
		}
		if (v=='%') {
			if(1!=fread(&v1,1,1,fd)) {
				return -1;
			}
			if (v1=='%') {
				*(p++)='%'; // % character
				continue;
			}
			if(1!=fread(&v2,1,1,fd)) {
				return -1;
			}
			if (v1>='A') {
				v1|=0x20;
				v1-='a';
				v1+=10;
			} else {
				v1-='0';
			}
			if (v1>15) {
				return -1;
			}
			if (v2>='A') {
				v2|=0x20;
				v2-='a';
				v2+=10;
			} else {
				v2-='0';
			}
			if (v2>15) {
				return -1;
			}
			v=v1*16+v2;
		}
		*(p++)=v;
	}
	return 0;
}

/**
 * Reloads a registry from disk, desotroying the changes done in memory. Useful to recover the original values if the user cancels a modification.
 *
 * @param object The registry to reload.
 *
 * @return 0 if all went fine, -1 if it was unable to open the registry file
 */

int32_t ttos_reload_registry(struct ttos_registry *object) {

	FILE *fd;
	Cstring final_path,base_path,tmp_path;
	struct ttos_registry_element *element;
	Cstring key,key2;
	uint8_t type;
	long long unsigned int value;

	minibus_list_free(&object->elements);

	tmp_path=getenv("TTOS_BASE_PATH");
	if ((!tmp_path)||(!tmp_path[0])) {
		final_path=minibus_path_join("reg/",object->name);
	} else {
		base_path=minibus_path_join(tmp_path,"reg");
		final_path=minibus_path_join(base_path,object->name);
		free(base_path);
	}

	fd=fopen(final_path,"rb");
	free(final_path);
	if (fd==NULL) {
		return -1;
	}

	key2=malloc(TTOS_MAX_KEY_SIZE+1);

	while(!feof(fd)) {
		if (3!=fscanf(fd,"%s %c %llu",key2,&type,&value)) {
			ttos_jump_nl(fd);
			continue;
		}
		key=strdup(key2);
		element=malloc(sizeof(struct ttos_registry_element));
		if (!element) {
			free (key);
			break;
		}
		element->key=key;
		switch (type) {
		case 'b': // byte
			if (value > 0x00FF) {
				free (element);
				element=NULL;
				free (key);
			} else {
				element->value.type=MINIBUS_TYPE_INT8;
				element->value.value8=(uint8_t) value;
			}
		break;
		case 's': // short
			if (value > 0x00FFFF) {
				free (element);
				element=NULL;
				free (key);
			} else {
				element->value.type=MINIBUS_TYPE_INT16;
				element->value.value16=(uint16_t) value;
			}
		break;
		case 'l': // long
			if (value > 0x00FFFFFFFF) {
				free (element);
				element=NULL;
				free (key);
			} else {
				element->value.type=MINIBUS_TYPE_INT32;
				element->value.value32=(uint32_t) value;
			}
		break;
		case 'c': // character string
			if (value > 0x00FFFF) {
				free(element);
				element=NULL;
				free (key);
			} else {
				element->value.type=MINIBUS_TYPE_STRING;
				element->value.stringsize=(uint16_t) (value-1);
				element->value.string=malloc(value);
				if (!element->value.string) {
					free(element);
					element=NULL;
					free (key);
				} else {
					if (-1==ttos_read_blob(fd,(uint8_t *)element->value.string,value)) {
						free(element);
						element=NULL;
						free (key);
					}
				}
			}
		break;
		case 'r': // BLOB or RAW format
			if (value > 0x00FFFF) {
				free(element);
				element=NULL;
				free (key);
			} else {
				element->value.type=MINIBUS_TYPE_BLOB;
				element->value.blobsize=(uint16_t) value;
				element->value.blob=malloc(value);
				if (!element->value.blob) {
					free(element);
					element=NULL;
					free (key);
				} else {
					if (-1==ttos_read_blob(fd,element->value.blob,value)) {
						free(element);
						element=NULL;
						free (key);
					}
				}
			}
		break;
		}
		ttos_jump_nl (fd);
		if (element) {
			minibus_list_add(&object->elements,element);
		}
	}
	free(key2);
	fclose(fd);
	return 0;
}


/**
 * Creates a new registry object with the specified name, whose values will be stored in $TTOS_BASE_PATH/reg/name. If the filename
 * already exists, the pairs key-value in that file will be loaded into the registry object
 *
 * @param name The filename where to store the values
 * @return A registry object, or NULL if failed.
 */

struct ttos_registry *ttos_new_registry(Cstring name) {

	struct ttos_registry *object;

	object=malloc(sizeof(struct ttos_registry));

	if (!object) {
		return NULL;
	}
	object->name=strdup(name);
	minibus_list_init(&object->elements,ttos_destroy_registry_element,NULL);
	ttos_reload_registry(object);
	return (object);
}

/**
 * Destroys a registry object. It doesn't store the pairs key-value at disk, so the program must call before to ttos_store_registry
 *
 * @param object The registry object to destroy
 */

void ttos_destroy_registry(struct ttos_registry *object) {

	if (!object) {
		return;
	}
	free(object->name);
	minibus_list_free(&object->elements);
	free(object);
}

/**
 * Removes a key-value pair from the specified registry object.
 *
 * @param registry The registry object
 * @param key The key to remove
 *
 * @return 0 if all went fine, -1 if the registry or the key values were no valid
 */

int32_t ttos_registry_remove_key(struct ttos_registry *registry, Cstring key) {

	struct minibus_list_element *element;
	struct ttos_registry_element *relement;

	if ((!key)||(!registry)) {
		return -1;
	}

	for(element=registry->elements.list;element;element=element->next) {
		relement=element->data;
		if (!strcmp(key,relement->key)) {
			minibus_list_remove(element,true);
			break;
		}
	}
	return 0;
}

/**
 * Inserts a new key-value pair in a registry. The value can be a 8, 16 or 32bit integer, an string or a blob
 *
 * @param registry The registry object into which insert the key-value pair
 * @param key The key name
 * @param value The value. It must have the 'type' field filled with the data type
 *
 * @return 0 if all went fine; -1 if there was a memory problem; -2 if the key contains an invalid character; -3 if the key size is too long
 */

int32_t ttos_registry_insert_key(struct ttos_registry *registry, Cstring key, struct minibus_data_element *value) {

	struct ttos_registry_element *relement;
	Cstring p;

	if ((!key)||(!registry)||(!value)) {
		return -1;
	}

	if (strlen(key)>TTOS_MAX_KEY_SIZE) {
		return -3; // key too big
	}

	for(p=key;*p;p++) {
		if (*p<=' ') {
			return -2; // the key contains an invalid character
		}
	}

	// check if that key already exists and remove it
	ttos_registry_remove_key(registry,key);

	relement=malloc(sizeof(struct ttos_registry_element));
	if (!relement) {
		return -1;
	}
	relement->key=strdup(key);
	if (!relement->key) {
		free(relement);
		return -1;
	}
	relement->value.type=value->type;
	switch (value->type) {
	case MINIBUS_TYPE_INT8:
		relement->value.value8=value->value8;
	break;
	case MINIBUS_TYPE_INT16:
		relement->value.value16=value->value16;
	break;
	case MINIBUS_TYPE_INT32:
		relement->value.value32=value->value32;
	break;
	case MINIBUS_TYPE_STRING:
		relement->value.stringsize=strlen(value->string);
		relement->value.string=strdup(value->string);
		if (!relement->value.string) {
			free(relement->key);
			free(relement);
			return -1;
		}
	break;
	case MINIBUS_TYPE_BLOB:
		relement->value.blobsize=value->blobsize;
		relement->value.blob=malloc(value->blobsize);
		if (!relement->value.blob) {
			free(relement->key);
			free(relement);
			return -1;
		}
		memcpy(relement->value.blob,value->blob,value->blobsize);
	break;
	default:
		free(relement->key);
		free(relement);
		return -1;
	break;
	}
	minibus_list_add(&registry->elements,relement);
	return 0;
}


/**
 * Returns de value associated to a key
 *
 * @param registry The registry from which retrieve the value
 * @param key The key
 *
 * @return The value associated to the key, or NULL if there was a problem
 */

struct ttos_registry_element * ttos_registry_retrieve_key(struct ttos_registry *registry, Cstring key) {

	struct minibus_list_element *element;
	struct ttos_registry_element *relement;

	if ((!key)||(!registry)) {
		return NULL;
	}

	for(element=registry->elements.list;element;element=element->next) {
		relement=element->data;
		if (!strcmp(key,relement->key)) {
			return (relement);
		}
	}
	return NULL;
}

void ttos_fprint_blob(FILE *fd,uint8_t *buffer, uint32_t size) {

	uint32_t loop,value;
	uint8_t *p;

	p=buffer;
	for(loop=0;loop<size;loop++) {
		if (*p<' ') {
			value=*(p++);
			fprintf(fd,"%%%02X",value);
			continue;
		}
		if (*p=='%') {
			fprintf(fd,"%%%%");
			p++;
			continue;
		}
		fprintf(fd,"%c",*(p++));
	}
}


/**
 * Stores a registry object in disk, overwritting the old values
 *
 * @param object The registry object to save
 *
 * @return 0 if all went fine; -1 if there was an error
 */

int32_t ttos_store_registry(struct ttos_registry *object) {

	FILE *fd;
	Cstring final_path,base_path,tmp_path;
	struct ttos_registry_element *relement;
	struct minibus_list_element *element;

	tmp_path=getenv("TTOS_BASE_PATH");
	if ((!tmp_path)||(!tmp_path[0])) {
		mkdir("reg",0777);
		final_path=minibus_path_join("reg/",object->name);
	} else {
		base_path=minibus_path_join(tmp_path,"reg");
		mkdir(base_path,0777);
		final_path=minibus_path_join(base_path,object->name);
		free(base_path);
	}

	fd=fopen(final_path,"wb");
	free(final_path);
	if (!fd) {
		return -1;
	}

	for(element=object->elements.list;element;element=element->next) {
		relement=element->data;
		fprintf (fd,"%s ",relement->key);
		switch(relement->value.type) {
		case MINIBUS_TYPE_INT8:
			fprintf(fd,"b %u\n",relement->value.value8);
		break;
		case MINIBUS_TYPE_INT16:
			fprintf(fd,"s %u\n",relement->value.value16);
		break;
		case MINIBUS_TYPE_INT32:
			fprintf(fd,"l %u\n",relement->value.value32);
		break;
		case MINIBUS_TYPE_STRING:
			fprintf(fd,"c %d ",relement->value.stringsize+1);
			ttos_fprint_blob(fd,(uint8_t *)relement->value.string,relement->value.stringsize+1);
			fprintf(fd,"\n");
		break;
		case MINIBUS_TYPE_BLOB:
			fprintf(fd,"r %d ",relement->value.blobsize);
			ttos_fprint_blob(fd,(uint8_t *)relement->value.blob,relement->value.blobsize);
			fprintf(fd,"\n");
		break;
		default:
		break;
		}
	}
	fclose(fd);
	return 0;
}
