/*
 * Copyright 2011 Raster Software Vigo
 *
 * This file is part of ttos
 *
 * ttos is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * ttos is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

/*
 * widgets.c
 *
 *  Created on: 25/09/2010
 *      Author: raster
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/mman.h>
#include <unistd.h>

#include "other.h"
#include "widgets.h"
#include "uid.h"


bool widgets_event_cb(struct ttos_tssample *sample, struct _widget *object) {

	struct minibus_connection *conn;
	enum TTOS_TOUCH_SCREEN_EVENTS event;

	if (!object->event_port) {
		return false;
	}

	switch(sample->type) {
	case TTOS_TOUCH_CLICK:
		if (!object->click_sensitive)
			return false;
		event=TTOS_TS_CLICK;
	break;
	case TTOS_TOUCH_DRAG:
		if (!object->drag_sensitive)
			return false;
		event=TTOS_TS_DRAG;
	break;
	case TTOS_TOUCH_END_DRAG:
		if (!object->drag_sensitive)
			return false;
		event=TTOS_TS_END_DRAG;
	break;
	default:
		return false;
	break;
	}

	conn=object->parent_tab->client->conn;

	minibus_start_message(write_conn,object->event_port,TTOS_NO_ANSWER);
	minibus_write_long(write_conn,TTOS_SR_TOUCH_EVENT);
	minibus_write_byte(write_conn,event);
	minibus_write_long(write_conn,object->widget_id);
	minibus_write_long(write_conn,sample->x);
	minibus_write_long(write_conn,sample->y);
	minibus_write_long(write_conn,sample->ox);
	minibus_write_long(write_conn,sample->oy);
	minibus_server_send(write_conn,conn,false);

	return true;
}


/**
 * Initializes a base widget
 *
 * @param object The base widget to initializate
 * @param parent_tab The parent tab to which this widget belongs
 * @param widget_id The ID value for this widget
 * @param type The widget type
 */

void widgets_init_base_widget(struct _widget *object, struct _tab *parent_tab, uint32_t widget_id,enum _widget_type type) {

	if (!object) {
		return;
	}

	object->widget_id=widget_id;
	object->type=type;
	object->parent_tab=parent_tab;
	object->event_cb=NULL;
	object->repaint_callback=NULL;
	object->visible=true;
	object->h_fixed=false;
	object->v_fixed=false;
	object->enabled=false;
	object->ox=0;
	object->oy=0;
	object->w=0;
	object->h=0;

}


/**
 * Returns a pointer to a widget based on its id
 *
 * @param tab The tab which contains the widget
 * @param widget_id The id to find
 *
 * @return A pointer to the widget, or NULL if the widget doesn't exists
 */

struct _widget *widgets_find_widget(struct _tab *tab, uint32_t widget_id) {

	struct minibus_list_element *element;
	struct _widget *widget;

	if (!tab) {
		return NULL;
	}

	for(element=tab->widgets.list;element!=NULL;element=element->next) {
		widget=element->data;
		if (widget->widget_id==widget_id) {
			return widget;
		}
	}
	return NULL;
}


/**
 * Sets the visibility status for a widget
 *
 */

uint32_t widgets_set_visibility(uint32_t port, uint32_t answer_port, struct minibus_connection *connection, void *private_data) {

	uint32_t widget_id;
	uint16_t tab_id;
	bool visible;
	int32_t error;
	struct _widget *object;
	struct minibus_data_element *data;
	struct _client_data *client;

	client=connection->data;

	data=minibus_read_params(connection,"lb",&error);

	if(!data) {
		switch(error) {
		case -1:
			other_send_error(connection,answer_port,TTOS_ERROR_NO_MEMORY);
			return -1;
		break;
		case -2:
			other_send_error(connection,answer_port,TTOS_ERROR_CANT_READ_DATA);
			return -1;
		break;
		case -3:
			other_send_error(connection,answer_port,TTOS_ERROR_INCORRECT_FORMAT);
			return -1;
		break;
		}
	}

	widget_id=data[0].value32;
	visible=data[1].value8;
	free(data);

	if ((widget_id&0xF0000000)!=TTOS_WIDGET_UID_MASK) {
		other_send_error(connection,answer_port,TTOS_ERROR_INVALID_UID);
		return (-1);
	}

	tab_id=(uint16_t)((widget_id>>16)&0x0FFF);
	if (false==tabs_select_edit_tab(connection,tab_id)) {
		other_send_error(connection,answer_port,TTOS_ERROR_INVALID_UID);
		return (-1);
	}

	object=widgets_find_widget(client->edit_tab,widget_id);
	if (!object) {
		other_send_error(connection,answer_port,TTOS_ERROR_INVALID_UID);
		return (-1);
	}
	object->visible=visible;
	other_send_error(connection,answer_port,TTOS_ERROR_NO_ERROR);
	return 0;
}

/**
 * Changes a widget's position
 *
 */

uint32_t widgets_set_position(uint32_t port, uint32_t answer_port, struct minibus_connection *connection, void *private_data) {

	uint32_t widget_id;
	uint16_t tab_id;
	int16_t x,ox;
	int16_t y,oy;
	int32_t error;
	struct _widget *object;
	struct _label_widget *label;
	struct minibus_data_element *data;
	struct _client_data *client;

	client=connection->data;

	data=minibus_read_params(connection,"lss",&error);

	if(!data) {
		switch(error) {
		case -1:
			other_send_error(connection,answer_port,TTOS_ERROR_NO_MEMORY);
			return -1;
		break;
		case -2:
			other_send_error(connection,answer_port,TTOS_ERROR_CANT_READ_DATA);
			return -1;
		break;
		case -3:
			other_send_error(connection,answer_port,TTOS_ERROR_INCORRECT_FORMAT);
			return -1;
		break;
		}
	}

	widget_id=data[0].value32;
	x=data[1].value16;
	y=data[2].value16;
	free(data);

	if ((widget_id&0xF0000000)!=TTOS_WIDGET_UID_MASK) {
		other_send_error(connection,answer_port,TTOS_ERROR_INVALID_UID);
		return (-1);
	}

	tab_id=(uint16_t)((widget_id>>16)&0x0FFF);
	if (false==tabs_select_edit_tab(connection,tab_id)) {
		other_send_error(connection,answer_port,TTOS_ERROR_INVALID_UID);
		return (-1);
	}

	object=widgets_find_widget(client->edit_tab,widget_id);
	if (!object) {
		other_send_error(connection,answer_port,TTOS_ERROR_INVALID_UID);
		return (-1);
	}

	if (object->type==TTOS_WIDGET_LABEL) {
		label=(struct _label_widget *)object;
		ox=(object->ox-label->o_x);
		oy=(object->oy-label->o_y);
		label->o_x=x;
		label->o_y=y;
		x+=ox;
		y+=oy;
	}

	object->ox=x;
	object->oy=y;
	other_send_error(connection,answer_port,TTOS_ERROR_NO_ERROR);
	return 0;
}



void widgets_repaint_picture_cb (struct _widget *widget, int16_t x_offset, int16_t y_offset, struct _window *parent_window) {

	struct _picture_widget *picture=(struct _picture_widget *)widget;
	TTOS_Surface *bitmap=picture->bitmap;
	blit_blend(bitmap,0,0,bitmap->w,bitmap->h,base,x_offset,y_offset,parent_window->ox,parent_window->oy,parent_window->w,parent_window->h,false);

}


/**
 * Creates a new picture widget. Receives a message with the following fields:
 *
 * int16 tab_id The tab where the widget will be added
 * int16 x The X coordinate where the picture will be put
 * int16 y The Y coordinate where the picture will be put
 * int32 flags
 * int32 event port The port into which receive events
 * string The file name for the picture
 */

uint32_t widgets_create_picture(uint32_t port, uint32_t answer_port, struct minibus_connection *connection, void *private_data) {

	uint16_t tab_id;
	uint32_t widget_id;
	int16_t x,y;
	uint32_t mask;
	Cstring filename;
	int32_t error;
	uint32_t event_port;
	struct _picture_widget *object;
	struct minibus_data_element *data;
	struct _client_data *client;

	client=connection->data;

	data=minibus_read_params(connection,"sssllc",&error);

	if(!data) {
		switch(error) {
		case -1:
			other_send_error(connection,answer_port,TTOS_ERROR_NO_MEMORY);
			return -1;
		break;
		case -2:
			other_send_error(connection,answer_port,TTOS_ERROR_CANT_READ_DATA);
			return -1;
		break;
		case -3:
			other_send_error(connection,answer_port,TTOS_ERROR_INCORRECT_FORMAT);
			return -1;
		break;
		}
	}

	tab_id=data[0].value16;
	x=data[1].value16;
	y=data[2].value16;
	mask=data[3].value32;
	event_port=data[4].value32;
	filename=strdup(data[5].string);
	free(data);

	if (false==tabs_select_edit_tab(connection,tab_id)) {
		other_send_error(connection,answer_port,TTOS_ERROR_INVALID_UID);
		free(filename);
		return (-1);
	}

	// get an UID for this widget
	widget_id=uid_mix_uid(client->edit_tab->widget_id_counter,tab_id,TTOS_WIDGET_UID_MASK);
	client->edit_tab->widget_id_counter++;

	if(widgets_find_widget(client->edit_tab,widget_id)) {
		// we can't make more widgets (Hey, the limit is 65536 per tab!)
		other_send_error(connection,answer_port, TTOS_ERROR_TOO_MANY_WIDGETS);
		free(filename);
		return (-1);
	}

	object=malloc(sizeof(struct _picture_widget));
	if (!object) {
		other_send_error(connection,answer_port,TTOS_ERROR_NO_MEMORY);
		free(filename);
		return (-1);
	}
	widgets_init_base_widget((struct _widget *)object,client->edit_tab,widget_id,TTOS_WIDGET_PICTURE);

	object->h_fixed=(mask&TTOS_WIDGET_H_FIXED) ? true : false;
	object->v_fixed=(mask&TTOS_WIDGET_V_FIXED) ? true : false;
	object->click_sensitive=(mask&TTOS_WIDGET_CLICK_EVENT) ? true : false;
	object->drag_sensitive=(mask&TTOS_WIDGET_DRAG_EVENT) ? true : false;
	if ((object->click_sensitive)||(object->drag_sensitive)) {
		object->event_cb=widgets_event_cb;
		object->enabled=true;
	} else {
		object->event_cb=NULL;
		object->enabled=false;
	}
	object->event_port=event_port;

	object->bitmap=bitmaps_load(filename);
	if (!object->bitmap) {
		free(filename);
		free(object);
		other_send_error(connection,answer_port,TTOS_ERROR_FILE_NOT_FOUND);
		return (-1);
	}

	object->ox=x;
	object->oy=y;
	object->w=object->bitmap->w;
	object->h=object->bitmap->h;
	object->repaint_callback=widgets_repaint_picture_cb;

	minibus_list_add(&client->edit_tab->widgets,object);
	tabs_set_minmax(client->edit_tab,(struct _widget *)object);
	free(filename);

	other_send_uid(connection,answer_port,widget_id);
	return 0;
}


void widgets_repaint_drawable_cb (struct _widget *widget, int16_t x_offset, int16_t y_offset, struct _window *parent_window) {

	struct _drawable_widget *drawable=(struct _drawable_widget *)widget;
	TTOS_Surface *bitmap=drawable->bitmap;
	blit_blend(bitmap,0,0,bitmap->w,bitmap->h,base,x_offset,y_offset,parent_window->ox,parent_window->oy,parent_window->w,parent_window->h,false);

}

/**
 * Updates a drawable with the data sent. Receives a message with the following parameters
 * int32 drawable_id The drawable to update
 * int32 offset The offset of the zone to update (in pixels)
 * int32 flags Several flags
 * blob Picture data in RGB or RGBA format
 */

uint32_t widgets_update_drawable(uint32_t port, uint32_t answer_port, struct minibus_connection *connection, void *private_data) {

	uint32_t widget_id;
	uint32_t offset;
	uint32_t flags;
	bool has_alpha;
	int32_t error;
	uint8_t *blob;
	uint16_t size,counter,tab_id;
	uint32_t length,p_length;
	uint16_t r,g,b,a;
	uint16_t *pixel_data;
	uint8_t *alpha_data;
	struct _drawable_widget *object;
	struct minibus_data_element *data;
	struct _client_data *client;

	client=connection->data;

	data=minibus_read_params(connection,"llln",&error);

	if(!data) {
		switch(error) {
		case -1:
			other_send_error(connection,answer_port,TTOS_ERROR_NO_MEMORY);
			return -1;
		break;
		case -2:
			other_send_error(connection,answer_port,TTOS_ERROR_CANT_READ_DATA);
			return -1;
		break;
		case -3:
			other_send_error(connection,answer_port,TTOS_ERROR_INCORRECT_FORMAT);
			return -1;
		break;
		}
	}

	widget_id=data[0].value32;
	offset=data[1].value32;
	flags=data[2].value32;
	blob=data[3].blob;
	size=data[3].blobsize;

	has_alpha= (flags&TTOS_DRAWABLE_HAS_ALPHA) ? true : false;

	tab_id=(uint16_t)((widget_id>>16)&0x0FFF);
	if (false==tabs_select_edit_tab(connection,tab_id)) {
		other_send_error(connection,answer_port,TTOS_ERROR_INVALID_UID);
		return (-1);
	}

	object=(struct _drawable_widget *)widgets_find_widget(client->edit_tab,widget_id);
	if (!object) {
		free(data);
		other_send_error(connection,answer_port,TTOS_ERROR_INVALID_UID);
		return (-1);
	}

	if (has_alpha) {
		length=size/4;
	} else {
		length=size/3;
	}
	p_length=object->bitmap->w*object->bitmap->h;
	if (offset>=p_length) {
		free(data);
		other_send_error(connection,answer_port,TTOS_ERROR_OFFSET_OUT_OF_RANGE);
	}
	if ((length+offset)>p_length) {
		length=p_length-offset;
	}

	pixel_data=object->bitmap->pixels+offset;
	if (object->bitmap->alpha) {
		alpha_data=object->bitmap->alpha+offset;
	} else {
		alpha_data=NULL;
	}

	a=0x00;
	for(counter=0;counter<length;counter++) {
		r=(uint16_t)*((blob++));
		g=(uint16_t)*((blob++));
		b=(uint16_t)(*(blob++));
		if (has_alpha) {
			a=(uint16_t)(*(blob++));
			r*=a;
			g*=a;
			b*=a;
			a=~a;
			*(pixel_data++)=(r&0xF800)|((g>>5)&0x07E0)|((b>>11)&0x001F);
		} else {
			*(pixel_data++)=((r<<8)&0xF800)|((g<<3)&0x07E0)|((b>>3)&0x001F);
		}
		if (alpha_data) {
			*(alpha_data++)=a;
		}
	}
	free(data);
	other_send_error(connection,answer_port,TTOS_ERROR_NO_ERROR);
	return 0;
}


/**
 * Creates a new drawable widget. Receives a message with the following fields:
 *
 * int16 tab_id The tab where the widget will be added
 * int16 x The X coordinate where the drawable will be put
 * int16 y The Y coordinate where the drawable will be put
 * int16 w The width
 * int16 h The height
 * int32 share memory object UID if using shared memory, or 0 if not
 * int32 Bit mask
 * int32 Event port
  */

uint32_t widgets_create_drawable(uint32_t port, uint32_t answer_port, struct minibus_connection *connection, void *private_data) {

	uint16_t tab_id;
	uint32_t widget_id;
	int16_t x,y;
	uint16_t w,h;
	uint32_t mask;
	int32_t error;
	uint32_t event_port;
	uint32_t shared_uid;
	struct _shared_mem *shared;
	struct _drawable_widget *object;
	struct minibus_data_element *data;
	struct _client_data *client;
	struct minibus_list_element *element;

	client=connection->data;

	data=minibus_read_params(connection,"ssssslll",&error);

	if(!data) {
		switch(error) {
		case -1:
			other_send_error(connection,answer_port,TTOS_ERROR_NO_MEMORY);
			return -1;
		break;
		case -2:
			other_send_error(connection,answer_port,TTOS_ERROR_CANT_READ_DATA);
			return -1;
		break;
		case -3:
			other_send_error(connection,answer_port,TTOS_ERROR_INCORRECT_FORMAT);
			return -1;
		break;
		}
	}

	tab_id=data[0].value16;
	x=data[1].value16;
	y=data[2].value16;
	w=data[3].value16;
	h=data[4].value16;
	shared_uid=data[5].value32;
	mask=data[6].value32;
	event_port=data[7].value32;
	free(data);

	if (false==tabs_select_edit_tab(connection,tab_id)) {
		other_send_error(connection,answer_port,TTOS_ERROR_INVALID_UID);
		return (-1);
	}

	// get an UID for this widget
	widget_id=uid_mix_uid(client->edit_tab->widget_id_counter,tab_id,TTOS_WIDGET_UID_MASK);
	client->edit_tab->widget_id_counter++;

	if(widgets_find_widget(client->edit_tab,widget_id)) {
		// we can't make more widgets (Hey, the limit is 65536 per tab!)
		other_send_error(connection,answer_port, TTOS_ERROR_TOO_MANY_WIDGETS);
		return (-1);
	}

	object=malloc(sizeof(struct _drawable_widget));
	if (!object) {
		other_send_error(connection,answer_port,TTOS_ERROR_NO_MEMORY);
		return (-1);
	}
	widgets_init_base_widget((struct _widget *)object,client->edit_tab,widget_id,TTOS_WIDGET_DRAWABLE);

	object->h_fixed=(mask&TTOS_WIDGET_H_FIXED) ? true : false;
	object->v_fixed=(mask&TTOS_WIDGET_V_FIXED) ? true : false;
	object->click_sensitive=(mask&TTOS_WIDGET_CLICK_EVENT) ? true : false;
	object->drag_sensitive=(mask&TTOS_WIDGET_DRAG_EVENT) ? true : false;
	if ((object->click_sensitive)||(object->drag_sensitive)) {
		object->event_cb=widgets_event_cb;
		object->enabled=true;
	} else {
		object->event_cb=NULL;
		object->enabled=false;
	}
	object->event_port=event_port;

	if (shared_uid) {
		for(element=client->sharemem.list;element;element=element->next) {
			shared=element->data;
			if (shared->uid==shared_uid) {
				break;
			} else {
				shared=NULL;
			}
		}
		if (!shared) {
			free(object);
			other_send_error(connection,answer_port,TTOS_ERROR_INVALID_SHAREMEM_UID);
			return (-1);
		}
		if ((!shared->has_alpha)&&(mask&TTOS_DRAWABLE_HAS_ALPHA)) {
			free(object);
			other_send_error(connection,answer_port, TTOS_ERROR_SHAREMEM_HAS_NO_ALPHA);
			return (-1);
		}

		object->bitmap=blit_create_surface_from_shared(shared,(mask&TTOS_DRAWABLE_HAS_ALPHA) ? true : false);
	} else {
		object->bitmap=blit_create_surface(w,h,(mask&TTOS_DRAWABLE_HAS_ALPHA) ? true : false);
	}
	if (!object->bitmap) {
		free(object);
		other_send_error(connection,answer_port,TTOS_ERROR_INVALID_SHAREMEM_UID);
		return (-1);
	}

	memset(object->bitmap->pixels,0,w*h*2);
	if (object->bitmap->alpha) {
		memset(object->bitmap->alpha,0,w*h);
	}
	object->ox=x;
	object->oy=y;
	object->w=w;
	object->h=h;
	object->repaint_callback=widgets_repaint_drawable_cb;

	minibus_list_add(&client->edit_tab->widgets,object);
	tabs_set_minmax(client->edit_tab,(struct _widget *)object);

	other_send_uid(connection,answer_port,widget_id);
	return 0;
}



/**
 * Frees the memory used by a widget
 *
 * @param widget The widget to destroy
 */

void widgets_free_widget(void *object) {

	struct _widget *widget=object;
	struct _picture_widget *picture;
	struct _label_widget *label;
	struct _button_widget *button;
	struct _area_widget *area;
	struct _drawable_widget *drawable;

	if (!object) {
		return;
	}

	switch(widget->type) {
	case TTOS_WIDGET_PICTURE:
		picture=object;
		bitmaps_unload(picture->bitmap);
		free(picture);
	break;
	case TTOS_WIDGET_LABEL:
		label=object;
		blit_destroy_surface(label->bitmap);
		free(label);
	break;
	case TTOS_WIDGET_BUTTON:
		button=object;
		blit_destroy_surface(button->bitmap);
		free(button);
	break;
	case TTOS_WIDGET_AREA:
		area=object;
		free(area);
	break;
	case TTOS_WIDGET_DRAWABLE:
		drawable=object;
		blit_destroy_surface(drawable->bitmap);
		free(drawable);
	break;
	}
}


void widgets_repaint_label_cb (struct _widget *widget, int16_t x_offset, int16_t y_offset, struct _window *parent_window) {

	struct _label_widget *label=(struct _label_widget *)widget;
	TTOS_Surface *bitmap=label->bitmap;
	blit_blend(bitmap,label->px_x,label->px_y,label->w,label->h,base,x_offset,y_offset,parent_window->ox,parent_window->oy,parent_window->w,parent_window->h,false);
}


/**
 * Renders a new text in a label
 *
 * @param object The label widget
 * @param text The text itself
 * @param mask The options for text rendering
 *
 * @return 0 if all went fine, -1 if there was an error
 */

int widgets_render_label_text(struct _label_widget *object,Cstring text, uint32_t mask) {

	int16_t xoffset,yoffset;

	if (object->bitmap) {
		blit_destroy_surface(object->bitmap);
		object->bitmap=NULL;
	}

	object->px_x=0;
	object->px_y=0;

	object->bitmap=fonts_render_label(base_font,text,mask,(mask&TTOS_WIDGET_FONT_MULTILINE) ? object->o_w : 0, &object->o_color);
	if (!object->bitmap) {
		return (-1);
	}

	xoffset=0;
	if (object->o_w>object->bitmap->w) {
		if (mask&TTOS_WIDGET_FONT_HALIGN_CENTER) {
			xoffset=(object->o_w-object->bitmap->w)/2;
		} else if (mask&TTOS_WIDGET_FONT_HALIGN_RIGHT) {
			xoffset=(object->o_w-object->bitmap->w);
		}
	}

	yoffset=(object->o_h-object->bitmap->h)/2;
	if (object->o_h>object->bitmap->h) {
		if (mask&TTOS_WIDGET_FONT_VALIGN_TOP) {
			yoffset=0;
		} else if (mask&TTOS_WIDGET_FONT_VALIGN_BOTTOM) {
			yoffset=(object->o_h-object->bitmap->h);
		}
	} else {
		if (mask&TTOS_WIDGET_FONT_MULTILINE) {
			yoffset=0;
		}
	}

	if ((!(mask&TTOS_WIDGET_FONT_MULTILINE))&& (mask&TTOS_WIDGET_FONT_BIG_LABEL_ALIGN_RIGHT)) {
		if (object->o_w<object->bitmap->w) {
			object->px_x=object->bitmap->w-object->o_w;
		}
	}

	object->ox=object->o_x+xoffset;
	object->oy=object->o_y+yoffset;
	if (object->o_w>object->bitmap->w) {
		object->w=object->bitmap->w;
	} else {
		object->w=object->o_w;
	}
	if (object->o_h>object->bitmap->h) {
		object->h=object->bitmap->h;
	} else {
		if (mask&(TTOS_WIDGET_FONT_UPDATE_SIZE|TTOS_WIDGET_FONT_MULTILINE)) {
			object->h=object->bitmap->h;
		} else {
			object->h=object->o_h;
		}
	}

	return 0;
}


/**
 * Changes the text in a label widget. Receives a message with the following fields:
 *
 * int32 widget_id The id for the new widget
 * int32 Bit mask: WIDGETS defines
 * string The text for the label
 */

uint32_t widgets_set_label_text(uint32_t port, uint32_t answer_port, struct minibus_connection *connection, void *private_data) {

	uint32_t widget_id;
	uint32_t mask;
	Cstring text;
	int32_t error;
	struct _label_widget *object;
	struct minibus_data_element *data;
	struct _client_data *client;

	client=connection->data;

	data=minibus_read_params(connection,"llc",&error);

	if(!data) {
		switch(error) {
		case -1:
			other_send_error(connection,answer_port,TTOS_ERROR_NO_MEMORY);
			return -1;
		break;
		case -2:
			other_send_error(connection,answer_port,TTOS_ERROR_CANT_READ_DATA);
			return -1;
		break;
		case -3:
			other_send_error(connection,answer_port,TTOS_ERROR_INCORRECT_FORMAT);
			return -1;
		break;
		}
	}

	widget_id=data[0].value32;
	mask=data[1].value32;
	text=strdup(data[2].string);
	free(data);

	if ((widget_id&0xF0000000)!=TTOS_WIDGET_UID_MASK) {
		other_send_error(connection,answer_port,TTOS_ERROR_INVALID_UID);
		free(text);
		return (-1);
	}

	if (false==tabs_select_edit_tab(connection,(uint16_t)((widget_id>>16)&0x0FFF))) {
		other_send_error(connection,answer_port,TTOS_ERROR_INVALID_UID);
		free(text);
		return (-1);
	}

	object=(struct _label_widget *)widgets_find_widget(client->edit_tab,widget_id);
	if (!object) {
		other_send_error(connection,answer_port,TTOS_ERROR_INVALID_UID);
		free(text);
		return (-1);
	}
	if (object->type!=TTOS_WIDGET_LABEL) {
		other_send_error(connection,answer_port,TTOS_ERROR_INVALID_UID);
		free(text);
		return (-1);
	}
	if (0==widgets_render_label_text(object,text,mask)) {
		free(text);
		other_send_error(connection,answer_port,TTOS_ERROR_NO_ERROR);
		return 0;
	} else {
		free(text);
		other_send_error(connection,answer_port,TTOS_ERROR_NO_MEMORY);
		return -1;
	}
}


/**
 * Changes the picture in a picture widget. Receives a message with the following fields:
 *
 * int16 widget_id The widget's uid
 * string The file name for the new picture
 */

uint32_t widgets_set_picture(uint32_t port, uint32_t answer_port, struct minibus_connection *connection, void *private_data) {

	uint32_t widget_id;
	Cstring filename;
	int32_t error;
	struct _picture_widget *object;
	struct minibus_data_element *data;
	struct _client_data *client;
	TTOS_Surface *bitmap;

	client=connection->data;

	data=minibus_read_params(connection,"lc",&error);

	if(!data) {
		switch(error) {
		case -1:
			other_send_error(connection,answer_port,TTOS_ERROR_NO_MEMORY);
			return -1;
		break;
		case -2:
			other_send_error(connection,answer_port,TTOS_ERROR_CANT_READ_DATA);
			return -1;
		break;
		case -3:
			other_send_error(connection,answer_port,TTOS_ERROR_INCORRECT_FORMAT);
			return -1;
		break;
		}
	}

	widget_id=data[0].value32;
	filename=strdup(data[1].string);
	free(data);

	if ((widget_id&0xF0000000)!=TTOS_WIDGET_UID_MASK) {
		other_send_error(connection,answer_port,TTOS_ERROR_INVALID_UID);
		free(filename);
		return (-1);
	}

	if (false==tabs_select_edit_tab(connection,(uint16_t)((widget_id>>16)&0x0FFF))) {
		other_send_error(connection,answer_port,TTOS_ERROR_INVALID_UID);
		free(filename);
		return (-1);
	}

	object=(struct _picture_widget *)widgets_find_widget(client->edit_tab,widget_id);
	if (!object) {
		other_send_error(connection,answer_port,TTOS_ERROR_INVALID_UID);
		free(filename);
		return (-1);
	}
	if (object->type!=TTOS_WIDGET_PICTURE) {
		other_send_error(connection,answer_port,TTOS_ERROR_INVALID_UID);
		free(filename);
		return (-1);
	}

	bitmap=bitmaps_load(filename);
	free(filename);
	if (!bitmap) {
		other_send_error(connection,answer_port,TTOS_ERROR_FILE_NOT_FOUND);
		return (-1);
	}
	bitmaps_unload(object->bitmap);
	object->bitmap=bitmap;
	object->w=bitmap->w;
	object->h=bitmap->h;
	other_send_error(connection,answer_port,TTOS_ERROR_NO_ERROR);
	return 0;
}


/**
 * Creates a new label widget. Receives a message with the following fields:
 *
 * int16 tab_id The tab where the widget will be added
 * int16 x The X coordinate where the label will be put
 * int16 y The Y coordinate where the label will be put
 * int16 width The width of the label
 * int16 height The height of the label
 * int32 Bit mask: WIDGETS defines
 * int32 Event port
 * string The text for the label
 * byte Red color component
 * byte Green color component
 * byte Blue color component
 */

uint32_t widgets_create_label(uint32_t port, uint32_t answer_port, struct minibus_connection *connection, void *private_data) {

	uint16_t tab_id;
	uint32_t widget_id;
	int16_t x,y,width,height;
	uint32_t mask;
	Cstring text;
	int32_t error;
	uint32_t event_port;
	struct _label_widget *object;
	struct minibus_data_element *data;
	struct _client_data *client;
	struct ttos_color color;

	client=connection->data;

	data=minibus_read_params(connection,"sssssllcbbb",&error);

	if(!data) {
		switch(error) {
		case -1:
			other_send_error(connection,answer_port,TTOS_ERROR_NO_MEMORY);
			return -1;
		break;
		case -2:
			other_send_error(connection,answer_port,TTOS_ERROR_CANT_READ_DATA);
			return -1;
		break;
		case -3:
			other_send_error(connection,answer_port,TTOS_ERROR_INCORRECT_FORMAT);
			return -1;
		break;
		}
	}

	tab_id=data[0].value16;
	x=data[1].value16;
	y=data[2].value16;
	width=data[3].value16;
	height=data[4].value16;
	mask=data[5].value32;
	event_port=data[6].value32;
	text=strdup(data[7].string);
	color.r=data[8].value8;
	color.g=data[9].value8;
	color.b=data[10].value8;
	color.a=255;
	free(data);

	if (false==tabs_select_edit_tab(connection,tab_id)) {
		other_send_error(connection,answer_port,TTOS_ERROR_INVALID_UID);
		free(text);
		return (-1);
	}

	// get an UID for this widget
	widget_id=uid_mix_uid(client->edit_tab->widget_id_counter,tab_id,TTOS_WIDGET_UID_MASK);
	client->edit_tab->widget_id_counter++;

	if(widgets_find_widget(client->edit_tab,widget_id)) {
		// we can't make more widgets (Hey, the limit is 65536 per tab!)
		other_send_error(connection,answer_port, TTOS_ERROR_TOO_MANY_WIDGETS);
		free(text);
		return (-1);
	}

	object=malloc(sizeof(struct _label_widget));
	if (!object) {
		other_send_error(connection,answer_port,TTOS_ERROR_NO_MEMORY);
		free(text);
		return (-1);
	}
	widgets_init_base_widget((struct _widget *)object,client->edit_tab,widget_id,TTOS_WIDGET_LABEL);

	object->h_fixed=(mask&TTOS_WIDGET_H_FIXED) ? true : false;
	object->v_fixed=(mask&TTOS_WIDGET_V_FIXED) ? true : false;
	object->click_sensitive=(mask&TTOS_WIDGET_CLICK_EVENT) ? true : false;
	object->drag_sensitive=(mask&TTOS_WIDGET_DRAG_EVENT) ? true : false;
	if ((object->click_sensitive)||(object->drag_sensitive)) {
		object->event_cb=widgets_event_cb;
		object->enabled=true;
	} else {
		object->event_cb=NULL;
		object->enabled=false;
	}

	object->event_port=event_port;
	object->repaint_callback=widgets_repaint_label_cb;

	object->o_x=x;
	object->o_y=y;
	object->o_w=width;
	object->o_h=height;
	object->o_color.r=color.r;
	object->o_color.g=color.g;
	object->o_color.b=color.b;
	object->o_color.a=color.a;

	object->bitmap=NULL;
	if (0!=widgets_render_label_text(object,text,mask)) {
		free(text);
		free(object);
		other_send_error(connection,answer_port,TTOS_ERROR_NO_MEMORY);
		return (-1);
	}

	minibus_list_add(&client->edit_tab->widgets,object);
	tabs_set_minmax(client->edit_tab,(struct _widget *)object);
	free(text);

	other_send_uid(connection,answer_port,widget_id);
	return 0;
}




void widgets_repaint_color_area_cb (struct _widget *widget, int16_t x_offset, int16_t y_offset, struct _window *parent_window) {

	struct _area_widget *area=(struct _area_widget *)widget;

	blit_fill(base,x_offset,y_offset,area->w,area->h,&area->color,true);

	//blit_blend(bitmap,0,0,label->w,label->h,base,x_offset,y_offset,parent_window->ox,parent_window->oy,parent_window->w,parent_window->h,false);
}


/**
 * Creates a new color area widget. Receives a message with the following fields:
 *
 * int16 tab_id The tab where the widget will be added
 * int16 widget_id The id for the new widget
 * int16 x The X coordinate for the area
 * int16 y The Y coordinate for the area
 * int16 width The width of the area
 * int16 height The height of the area
 * int32 Bit mask: WIDGETS defines
 * int32 Event port
 * byte Red color component
 * byte Green color component
 * byte Blue color component
 * byte Alpha color component
 */

uint32_t widgets_create_color_area(uint32_t port, uint32_t answer_port, struct minibus_connection *connection, void *private_data) {

	uint16_t tab_id;
	uint32_t widget_id;
	int16_t x,y,width,height;
	uint32_t mask;
	int32_t error;
	uint32_t event_port;
	struct _area_widget *object;
	struct minibus_data_element *data;
	struct _client_data *client;
	uint8_t r,g,b,a;

	client=connection->data;

	data=minibus_read_params(connection,"sssssllbbbb",&error);

	if(!data) {
		switch(error) {
		case -1:
			other_send_error(connection,answer_port,TTOS_ERROR_NO_MEMORY);
			return -1;
		break;
		case -2:
			other_send_error(connection,answer_port,TTOS_ERROR_CANT_READ_DATA);
			return -1;
		break;
		case -3:
			other_send_error(connection,answer_port,TTOS_ERROR_INCORRECT_FORMAT);
			return -1;
		break;
		}
	}

	tab_id=data[0].value16;
	x=data[1].value16;
	y=data[2].value16;
	width=data[3].value16;
	height=data[4].value16;
	mask=data[5].value32;
	event_port=data[6].value32;
	r=data[7].value8;
	g=data[8].value8;
	b=data[9].value8;
	a=data[10].value8;;
	free(data);

	if (false==tabs_select_edit_tab(connection,tab_id)) {
		other_send_error(connection,answer_port,TTOS_ERROR_INVALID_UID);
		return (-1);
	}

	// get an UID for this widget
	widget_id=uid_mix_uid(client->edit_tab->widget_id_counter,tab_id,TTOS_WIDGET_UID_MASK);
	client->edit_tab->widget_id_counter++;

	if(widgets_find_widget(client->edit_tab,widget_id)) {
		// we can't make more widgets (Hey, the limit is 65536 per tab!)
		other_send_error(connection,answer_port, TTOS_ERROR_TOO_MANY_WIDGETS);
		return (-1);
	}

	object=malloc(sizeof(struct _area_widget));
	if (!object) {
		other_send_error(connection,answer_port,TTOS_ERROR_NO_MEMORY);
		return (-1);
	}
	widgets_init_base_widget((struct _widget *)object,client->edit_tab,widget_id,TTOS_WIDGET_AREA);

	object->h_fixed=(mask&TTOS_WIDGET_H_FIXED) ? true : false;
	object->v_fixed=(mask&TTOS_WIDGET_V_FIXED) ? true : false;
	object->click_sensitive=(mask&TTOS_WIDGET_CLICK_EVENT) ? true : false;
	object->drag_sensitive=(mask&TTOS_WIDGET_DRAG_EVENT) ? true : false;
	if ((object->click_sensitive)||(object->drag_sensitive)) {
		object->event_cb=widgets_event_cb;
		object->enabled=true;
	} else {
		object->event_cb=NULL;
		object->enabled=false;
	}
	object->event_port=event_port;

	object->ox=x;
	object->oy=y;
	object->w=width;
	object->h=height;
	object->repaint_callback=widgets_repaint_color_area_cb;

	object->color.r=r;
	object->color.g=g;
	object->color.b=b;
	object->color.a=a;

	minibus_list_add(&client->edit_tab->widgets,object);
	tabs_set_minmax(client->edit_tab,(struct _widget *)object);

	other_send_uid(connection,answer_port,widget_id);

	return 0;
}


void widgets_repaint_button_cb (struct _widget *widget, int16_t x_offset, int16_t y_offset, struct _window *parent_window) {

	struct _button_widget *button=(struct _button_widget *)widget;
	TTOS_Surface *bitmap=button->bitmap;
	blit_blend(bitmap,0,0,button->w,button->h,base,x_offset,y_offset,parent_window->ox,parent_window->oy,parent_window->w,parent_window->h,false);
}


/**
 * Creates a new button widget. Receives a message with the following fields:
 *
 * int16 tab_id The tab where the widget will be added
 * int16 widget_id The id for the new widget
 * int16 x The X coordinate where the button will be put
 * int16 y The Y coordinate where the button will be put
 * int16 width The width of the button
 * int16 height The height of the button
 * int32 Bit mask: WIDGETS defines
 * string The path for a picture to be put at left (empty for no picture)
 * string The text for the button
 * byte Red color component for text
 * byte Green color component for text
 * byte Blue color component for text
 * byte Red color component for background
 * byte Green color component for background
 * byte Blue color component for background
 * byte Repaint 1 to repaint the tab after creating the label
 */

uint32_t widgets_create_button(uint32_t port, uint32_t answer_port, struct minibus_connection *connection, void *private_data) {

	uint16_t tab_id;
	uint32_t widget_id;
	int16_t x,y,width,height;
	uint32_t mask;
	Cstring picture;
	Cstring text;
	int32_t error;
	int32_t p_x,crop_y,crop_h;
	uint32_t event_port;
	struct _button_widget *object;
	struct minibus_data_element *data;
	struct _client_data *client;
	TTOS_Surface *b_label;
	TTOS_Surface *b_picture;
	struct ttos_color tcolor,bgcolor,rcolor;

	client=connection->data;

	data=minibus_read_params(connection,"sssssllccbbbbbbb",&error);

	if(!data) {
		switch(error) {
		case -1:
			other_send_error(connection,answer_port,TTOS_ERROR_NO_MEMORY);
			return -1;
		break;
		case -2:
			other_send_error(connection,answer_port,TTOS_ERROR_CANT_READ_DATA);
			return -1;
		break;
		case -3:
			other_send_error(connection,answer_port,TTOS_ERROR_INCORRECT_FORMAT);
			return -1;
		break;
		}
	}

	tab_id=data[0].value16;
	x=data[1].value16;
	y=data[2].value16;
	width=data[3].value16;
	height=data[4].value16;
	mask=data[5].value32;
	event_port=data[6].value32;
	picture=strdup(data[7].string);
	text=strdup(data[8].string);
	tcolor.r=data[9].value8;
	tcolor.g=data[10].value8;
	tcolor.b=data[11].value8;
	bgcolor.r=data[12].value8;
	bgcolor.g=data[13].value8;
	bgcolor.b=data[14].value8;
	bgcolor.a=data[15].value8;
	free(data);

	if (false==tabs_select_edit_tab(connection,tab_id)) {
		other_send_error(connection,answer_port,TTOS_ERROR_INVALID_TAB_UID);
		free(text);
		free(picture);
		return (-1);
	}

	// get an UID for this widget
	widget_id=uid_mix_uid(client->edit_tab->widget_id_counter,tab_id,TTOS_WIDGET_UID_MASK);
	client->edit_tab->widget_id_counter++;

	if(widgets_find_widget(client->edit_tab,widget_id)) {
		// we can't make more widgets (Hey, the limit is 65536 per tab!)
		other_send_error(connection,answer_port, TTOS_ERROR_TOO_MANY_WIDGETS);
		free(text);
		free(picture);
		return (-1);
	}

	object=malloc(sizeof(struct _button_widget));
	if (!object) {
		other_send_error(connection,answer_port,TTOS_ERROR_NO_MEMORY);
		free(text);
		free(picture);
		return (-1);
	}
	widgets_init_base_widget((struct _widget *)object,client->edit_tab,widget_id,TTOS_WIDGET_BUTTON);

	object->h_fixed=(mask&TTOS_WIDGET_H_FIXED) ? true : false;
	object->v_fixed=(mask&TTOS_WIDGET_V_FIXED) ? true : false;
	object->click_sensitive=(mask&TTOS_WIDGET_CLICK_EVENT) ? true : false;
	object->drag_sensitive=(mask&TTOS_WIDGET_DRAG_EVENT) ? true : false;
	if ((object->click_sensitive)||(object->drag_sensitive)) {
		object->event_cb=widgets_event_cb;
		object->enabled=true;
	} else {
		object->event_cb=NULL;
		object->enabled=false;
	}
	object->event_port=event_port;

	object->bitmap=blit_create_surface(width,height,true);
	if (!object->bitmap) {
		free(text);
		free(picture);
		free(object);
		other_send_error(connection,answer_port,TTOS_ERROR_NO_MEMORY);
		return (-1);
	}
	blit_fill(object->bitmap,0,0,width,height,&bgcolor,false);

	other_apply_alpha(&bgcolor,&rcolor,0);
	if ((object->bitmap->w>2)&&(object->bitmap->h>2)) {
		blit_multipixel(object->bitmap,0,0,&rcolor);
		other_apply_alpha(&bgcolor,&rcolor,148);
		blit_multipixel(object->bitmap,1,0,&rcolor);
		blit_multipixel(object->bitmap,0,1,&rcolor);
		other_apply_alpha(&bgcolor,&rcolor,236);
		blit_multipixel(object->bitmap,2,0,&rcolor);
		blit_multipixel(object->bitmap,0,2,&rcolor);
	}

	if ((picture[0]!=0)&&(width>4)&&(height>4)) {
		b_picture=bitmaps_load(picture);
		if (!b_picture) {
			free(text);
			free(picture);
			widgets_free_widget(object);
			other_send_error(connection,answer_port,TTOS_ERROR_CANT_LOAD_PICTURE);
			return (-1);
		}
		p_x=4+b_picture->w;
	} else {
		b_picture=NULL;
		p_x=2;
	}

	if ((p_x<width-2)&&(text[0])) {
		b_label=fonts_render_label(base_font,text,mask, width-2 ,&tcolor);
		if (!object->bitmap) {
			free(text);
			free(picture);
			widgets_free_widget(object);
			other_send_error(connection,answer_port,TTOS_ERROR_NO_MEMORY);
			return (-1);
		}
		if ((2+b_label->h)>object->bitmap->h) {
			crop_y=2;
			crop_h=object->bitmap->h-4;
		} else {
			crop_h=b_label->h;
			crop_y=(object->bitmap->h-b_label->h)/2;
			if (mask&TTOS_WIDGET_FONT_VALIGN_TOP) {
				crop_y=2;
			} else if (mask&TTOS_WIDGET_FONT_VALIGN_BOTTOM) {
				crop_y=(object->bitmap->h-b_label->h-2);
			}
		}
		if ((mask&TTOS_WIDGET_FONT_HALIGN_CENTER)) {
			p_x+=(width-2-p_x-b_label->w)/2;
		} else if ((mask&TTOS_WIDGET_FONT_HALIGN_RIGHT)) {
			p_x+=(width-2-p_x-b_label->w);
		}
		blit_blend(b_label,0,0,b_label->w,b_label->h,object->bitmap,p_x,crop_y,2,crop_y,object->bitmap->w-4,crop_h,false);
		blit_destroy_surface(b_label);
	} else {
		if ((mask&TTOS_WIDGET_FONT_HALIGN_CENTER)) {
			p_x+=(width-p_x)/2;
		} else if ((mask&TTOS_WIDGET_FONT_HALIGN_RIGHT)) {
			p_x+=(width-p_x);
		}
	}

	if (b_picture) {
		if ((2+b_picture->h)>object->bitmap->h) {
			crop_y=2;
			crop_h=object->bitmap->h-4;
		} else {
			crop_h=object->bitmap->h;
			crop_y=(object->bitmap->h-b_picture->h)/2;
			if (mask&TTOS_WIDGET_FONT_VALIGN_TOP) {
				crop_y=2;
			} else if (mask&TTOS_WIDGET_FONT_VALIGN_BOTTOM) {
				crop_y=(object->bitmap->h-b_picture->h-2);
			}
		}
		blit_blend(b_picture,0,0,b_picture->w,b_picture->h,object->bitmap,p_x-b_picture->w-2,crop_y,2,crop_y,object->bitmap->w-4,crop_h,false);
		bitmaps_unload(b_picture);
	}

	object->ox=x;
	object->oy=y;
	object->w=width;
	object->h=height;
	object->repaint_callback=widgets_repaint_button_cb;

	minibus_list_add(&client->edit_tab->widgets,object);
	tabs_set_minmax(client->edit_tab,(struct _widget *)object);

	free(text);
	free(picture);

	other_send_uid(connection,answer_port,widget_id);
	return 0;
}
